from .optimizer import Optimizer
import nestle
import numpy as np
import time
from taurex.util.util import quantile_corner, \
                                recursively_save_dict_contents_to_output


class NestleOptimizer(Optimizer):
    """ An optimizer that uses the `nestle <http://kylebarbary.com/nestle/>`_ library
    to perform optimization. 

    Parameters
    ----------
    observed : :class:`~taurex.data.spectrum.spectrum.BaseSpectrum`, optional
        Observed spectrum to optimize to

    model : :class:`~taurex.model.model.ForwardModel`, optional
        Forward model to optimize

    num_live_points : int, optional
        Number of live points to use in sampling

    method : ``classic``, ``single`` or ``multi``
        Nested sampling method to use. ``classic`` uses MCMC exploration,
        ``single`` uses a single ellipsoid and ``multi`` uses multiple ellipsoids (similar to Multinest) 

    tol : float
        Evidence tolerance value to stop the fit. This is based on an estimate of the remaining prior volumes
        contribution to the evidence.

    sigma_fraction: float, optional
        Fraction of weights to use in computing the error. (Default: 0.1)

    """

    def __init__(self, observed=None, model=None, num_live_points=1500, method='multi', tol=0.5, sigma_fraction=0.1):
        super().__init__('Nestle', observed, model, sigma_fraction)
        self._nlive = int(num_live_points)    # number of live points
        self._method = method  # use MutliNest algorithm

        self._tol = tol       # the stopping criterion
        self._nestle_output = None

    @property
    def tolerance(self):
        return self._tol

    @tolerance.setter
    def tolerance(self, value):
        self._tol = value

    @property
    def numLivePoints(self):
        return self._nlive

    @numLivePoints.setter
    def numLivePoints(self, value):
        self._nlive = value

    def compute_fit(self):
        """

        Computes the fit using nestle

        """
        data = self._observed.spectrum
        datastd = self._observed.errorBar
        sqrtpi = np.sqrt(2*np.pi)

        def nestle_loglike(params):
            # log-likelihood function called by multinest
            fit_params_container = np.array(params)
            chi_t = self.chisq_trans(fit_params_container, data, datastd)
            loglike = -np.sum(np.log(datastd*sqrtpi)) - 0.5 * chi_t
            return loglike

        def nestle_uniform_prior(theta):
            # prior distributions called by multinest. Implements a uniform prior
            # converting parameters from normalised grid to uniform prior
            cube = []

            for idx, bounds in enumerate(self.fit_boundaries):
                bound_min, bound_max = bounds
                cube.append((theta[idx] * (bound_max-bound_min)) + bound_min)

            return tuple(cube)

        ndim = len(self.fitting_parameters)
        self.warning('Beginning fit......')
        ndims = ndim        # two parameters

        t0 = time.time()

        res = nestle.sample(nestle_loglike, nestle_uniform_prior, ndims, method='multi',
                            npoints=self.numLivePoints, dlogz=self.tolerance, callback=nestle.print_progress)
        t1 = time.time()

        timenestle = (t1-t0)

        print(res.summary())

        self.warning("Time taken to run 'Nestle' is %s seconds", timenestle)

        self.warning('Fit complete.....')

        self._nestle_output = self.store_nestle_output(res)

    def sample_parameters(self, solution):
        """
        Read traces and weights and return
        a random ``sigma_fraction`` sample of them

        Parameters
        ----------
        solution:
            a solution output from sampler

        Yields
        ------
        traces: :obj:`array`
            Traces of a particular sample

        weight: float
            Weight of sample
        
        """
        from taurex.util.util import random_int_iter
        samples = self._nestle_output['solution']['samples']
        weights = self._nestle_output['solution']['weights']

        for x in random_int_iter(samples.shape[0], self._sigma_fraction):
            w = weights[x]+1e-300

            yield samples[x, :], w

    def get_solution(self):
        """

        Generator for solutions and their
        median and MAP values

        Yields
        ------

        solution_no: int
            Solution number (always 0)

        map: :obj:`array`
            Map values

        median: :obj:`array`
            Median values

        extra: :obj:`list`
            Returns Statistics, fitting_params, raw_traces and
            raw_weights

        """

        names = self.fit_names
        opt_map = self.fit_values
        opt_values = self.fit_values
        for k, v in self._nestle_output['solution']['fitparams'].items():
            if k in ('mu_derived',):
                continue
            idx = names.index(k)
            opt_map[idx] = v['map']
            opt_values[idx] = v['value']

        yield 0, opt_map, opt_values, [('Statistics', self._nestle_output['Stats']),
                                       ('fit_params',
                                        self._nestle_output['solution']['fitparams']),
                                       ('tracedata',
                                        self._nestle_output['solution']['samples']),
                                       ('weights', self._nestle_output['solution']['weights'])]

    def write_optimizer(self, output):
        opt = super().write_optimizer(output)

        # number of live points
        opt.write_scalar('num_live_points', self._nlive)
        # maximum no. of iterations (0=inf)
        opt.write_string('method', self._method)
        # search for multiple modes
        opt.write_scalar('tol', self._tol)

        return opt

    def write_fit(self, output):
        fit = super().write_fit(output)

        if self._nestle_output:
            recursively_save_dict_contents_to_output(
                output, self._nestle_output)

        return fit

    def store_nestle_output(self, result):
        """
        This turns the output fron nestle into a dictionary that can
        be output by Taurex

        Parameters
        ----------
        result: :obj:`dict`
            Result from a nestle sample call

        Returns
        -------
        dict
            Formatted dictionary for output

        """

        from tabulate import tabulate

        nestle_output = {}
        nestle_output['Stats'] = {}
        nestle_output['Stats']['Log-Evidence'] = result.logz
        nestle_output['Stats']['Log-Evidence-Error'] = result.logzerr
        nestle_output['Stats']['Peakiness'] = result.h

        fit_param = self.fit_names

        samples = result.samples
        weights = result.weights

        mean, cov = nestle.mean_and_cov(samples, weights)
        nestle_output['solution'] = {}
        nestle_output['solution']['samples'] = samples
        nestle_output['solution']['weights'] = weights
        nestle_output['solution']['covariance'] = cov
        nestle_output['solution']['fitparams'] = {}

        max_weight = weights.argmax()

        table_data = []

        for idx, param_name in enumerate(fit_param):
            param = {}
            param['mean'] = mean[idx]
            param['sigma'] = cov[idx]
            trace = samples[:, idx]
            q_16, q_50, q_84 = quantile_corner(trace, [0.16, 0.5, 0.84],
                                               weights=np.asarray(weights))
            param['value'] = q_50
            param['sigma_m'] = q_50-q_16
            param['sigma_p'] = q_84-q_50
            param['trace'] = trace
            param['map'] = trace[max_weight]
            table_data.append((param_name, q_50, q_50-q_16))

            nestle_output['solution']['fitparams'][param_name] = param

        return nestle_output



from .tprofile import TemperatureProfile
import numpy as np
from taurex.data.fittable import fitparam
from taurex.util import movingaverage


class NPoint(TemperatureProfile):
    """

    A temperature profile that is defined at various heights of the
    atmopshere and then smoothend.

    At minimum, temepratures on both the top ``T_top`` and surface
    ``T_surface`` must be defined.
    If any intermediate points are given as ``temperature_points``
    then the same number of ``pressure_points``
    must be given as well.

    A 2-point temperature profile has ``len(temperature_points) == 0``
    A 3-point temperature profile has ``len(temperature_points) == 1``

    etc.


    Parameters
    -----------
        T_surface : float
            Temperature at the planets surface in Kelvin

        T_top : float
            Temperature at the top of the atmosphere in Kelvin

        P_surface : float , optional
            Pressure for ``T_surface`` (Optional) otherwise uses surface
            pressure from forward model

        P_top : float , optional
            Pressure for ``T_top`` (Optional) otherwise uses top pressure from
            forward model

        temperature_points : :obj:`list`
            temperature points between ``T_top`` and ``T_surface``

        pressure_points : :obj:`list`
            pressure points that the each temperature in ``temperature_points``
            lie on

        smoothing_window : int
            smoothing window


    """

    def __init__(self, T_surface=1500.0, T_top=200.0, P_surface=None,
                 P_top=None, temperature_points=[], pressure_points=[],
                 smoothing_window=10):
        super().__init__('{}Point'.format(len(temperature_points)+2))

        if not hasattr(temperature_points, '__len__'):
            raise Exception('t_point is not an iterable')

        if len(temperature_points) != len(pressure_points):
            self.error('Number of temeprature points != number of '
                       'pressure points')
            self.error('len(t_points) = %s /= '
                       'len(p_points) = %s',
                       len(temperature_points),
                       len(pressure_points))
            raise Exception('Incorrect_number of temp and pressure points')

        self.info('Npoint temeprature profile is initialized')
        self.debug('Passed temeprature points %s', temperature_points)
        self.debug('Passed pressure points %s', pressure_points)
        self._t_points = temperature_points
        self._p_points = pressure_points
        self._T_surface = T_surface
        self._T_top = T_top
        self._P_surface = P_surface
        self._P_top = P_top
        self._smooth_window = smoothing_window
        self.generate_pressure_fitting_params()
        self.generate_temperature_fitting_params()

    @fitparam(param_name='T_surface',
              param_latex='$T_\\mathrm{surf}$',
              default_fit=False,
              default_bounds=[300, 2500])
    def temperatureSurface(self):
        """Temperature at planet surface in Kelvin"""
        return self._T_surface

    @temperatureSurface.setter
    def temperatureSurface(self, value):
        self._T_surface = value

    @fitparam(param_name='T_top',
              param_latex='$T_\\mathrm{top}$',
              default_fit=False,
              default_bounds=[300, 2500])
    def temperatureTop(self):
        """Temperature at top of atmosphere in Kelvin"""
        return self._T_top

    @temperatureTop.setter
    def temperatureTop(self, value):
        self._T_top = value

    @fitparam(param_name='P_surface',
              param_latex='$P_\\mathrm{surf}$',
              default_fit=False,
              default_bounds=[1e3, 1e2],
              default_mode='log')
    def pressureSurface(self):
        return self._P_surface

    @pressureSurface.setter
    def pressureSurface(self, value):
        self._P_surface = value

    @fitparam(param_name='P_top',
              param_latex='$P_\\mathrm{top}$',
              default_fit=False,
              default_bounds=[1e-5, 1e-4],
              default_mode='log')
    def pressureTop(self):
        return self._P_top

    @pressureTop.setter
    def pressureTop(self, value):
        self._P_top = value

    def generate_pressure_fitting_params(self):
        """Generates the fitting parameters for the pressure points
        These are given the name ``P_point(number)`` for example, if two extra
        pressure points are defined between the top and surface then the
        fitting parameters generated are ``P_point0`` and ``P_point1``
        """

        bounds = [1e5, 1e3]
        for idx, val in enumerate(self._p_points):
            point_num = idx+1
            param_name = 'P_point{}'.format(point_num)
            param_latex = '$P_{}$'.format(point_num)

            def read_point(self, idx=idx):
                return self._p_points[idx]

            def write_point(self, value, idx=idx):
                self._p_points[idx] = value

            fget_point = read_point
            fset_point = write_point
            self.debug('FGet_location %s', fget_point)
            default_fit = False
            self.add_fittable_param(param_name, param_latex, fget_point,
                                    fset_point, 'log', default_fit, bounds)

    def generate_temperature_fitting_params(self):
        """Generates the fitting parameters for the temeprature points
        These are given the name ``T_point(number)`` for example, if two extra
        temeprature points are defined between the top and surface then the
        fitting parameters generated are ``T_point0`` and ``T_point1``
        """

        bounds = [300, 2500]
        for idx, val in enumerate(self._t_points):
            point_num = idx+1
            param_name = 'T_point{}'.format(point_num)
            param_latex = '$T_{}$'.format(point_num)

            def read_point(self, idx=idx):
                return self._t_points[idx]

            def write_point(self, value, idx=idx):
                self._t_points[idx] = value

            fget_point = read_point
            fset_point = write_point
            self.debug('FGet_location %s %s', fget_point, fget_point(self))
            default_fit = False
            self.add_fittable_param(param_name, param_latex, fget_point,
                                    fset_point, 'linear', default_fit, bounds)

    @property
    def profile(self):

        Tnodes = [self._T_surface, *self._t_points, self._T_top]

        Psurface = self._P_surface
        if Psurface is None or Psurface < 0:
            Psurface = self.pressure_profile[0]

        Ptop = self._P_top
        if Ptop is None or Ptop < 0:
            Ptop = self.pressure_profile[-1]

        Pnodes = [Psurface, *self._p_points, Ptop]

        smooth_window = self._smooth_window

        TP = np.interp((np.log(self.pressure_profile[::-1])),
                       np.log(Pnodes[::-1]), Tnodes[::-1])

        # smoothing T-P profile
        wsize = int(self.nlayers*(smooth_window / 100.0))
        if (wsize % 2 == 0):
            wsize += 1
        TP_smooth = movingaverage(TP, wsize)
        border = np.int((len(TP) - len(TP_smooth))/2)

        foo = TP[::-1]
        foo[border:-border] = TP_smooth[::-1]

        return foo

    def write(self, output):
        temperature = super().write(output)

        temperature.write_scalar('T_surface', self._T_surface)
        temperature.write_scalar('T_top', self._T_top)
        temperature.write_array('temperature_points', np.array(self._t_points))

        P_surface = self._P_surface
        P_top = self._P_top
        if not P_surface:
            P_surface = -1
        if not P_top:
            P_top = -1

        temperature.write_scalar('P_surface', P_surface)
        temperature.write_scalar('P_top', P_top)
        temperature.write_array('pressure_points', np.array(self._p_points))

        temperature.write_scalar('smoothing_window', self._smooth_window)

        return temperature



"""Module for wrapping MPI functions (future use)"""
from functools import lru_cache
from functools import wraps

@lru_cache(maxsize=2)
def nprocs():
    """Gets number of processes or returns 1 if mpi is not installed

    Returns
    -------
    int:
        Rank of process or 1 if MPI is not installed

    """
    try:
        from mpi4py import MPI
    except ImportError:
        return 1

    comm = MPI.COMM_WORLD

    return comm.Get_size()


def allgather(value):
    try:
        from mpi4py import MPI
    except ImportError:
        return [value]

    comm = MPI.COMM_WORLD
    data = value
    data = comm.allgather(data)

    return data


def broadcast(array, rank=0):
    import numpy as np
    try:
        from mpi4py import MPI
    except ImportError:
        return array
    comm = MPI.COMM_WORLD
    if isinstance(array, np.ndarray):

        data = None
        if get_rank() == rank:
            data = np.copy(array)
        else:
            data = np.zeros_like(array)
        comm.Bcast(data, root=rank)
    else:

        data = comm.bcast(array, root=rank)

    return data


@lru_cache(maxsize=2)
def get_rank():
    """Gets rank or returns 0 if mpi is not installed

    Returns
    -------
    int:
        Rank of process or 0 if MPI is not installed

    """

    try:
        from mpi4py import MPI
    except ImportError:
        return 0

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()

    return rank


def barrier():
    """Gets rank or returns 0 if mpi is not installed

    Returns
    -------
    int:
        Rank of process or 0 if MPI is not installed

    """

    try:
        from mpi4py import MPI
    except ImportError:
        return

    comm = MPI.COMM_WORLD
    comm.Barrier()


def only_master_rank(f):
    """
    A decorator to ensure only the master
    MPI rank can run it
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        if get_rank() == 0:
            return f(*args, **kwargs)
    return wrapper



from .gas import Gas
import numpy as np


class ArrayGas(Gas):
    """

    Gas profile from an array. Molecular abundance is interpolated if the
    number of layers do not match

    Parameters
    -----------
    molecule_name : str
        Name of molecule

    mix_ratio_array : :obj:`array`
        Mixing ratio of the molecule

    """

    def __init__(self, molecule_name='H2O',
                 mix_ratio_array=[1e-2, 1e-6]):
        super().__init__(self.__class__.__name__, molecule_name)

        self._mix_ratio_array = np.array(mix_ratio_array)

    @property
    def mixProfile(self):
        """

        Mixing profile

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer

        """

        return self._mix_array

    def initialize_profile(self, nlayers=None, temperature_profile=None,
                           pressure_profile=None, altitude_profile=None):

        if nlayers is None:
            self.error('number layers argument required')
            raise ValueError('Number of layers argument required')

        interp_array = np.linspace(0.0, 1.0, self._mix_ratio_array.shape[0])

        layer_interp = np.linspace(0.0, 1.0, nlayers)

        self._mix_array = np.interp(layer_interp,
                                    interp_array, self._mix_ratio_array)

    def write(self, output):
        """

        Parameters
        ----------
        output: :class:`~taurex.output.output.Output`
        """
        gas_entry = super().write(output)
        gas_entry.write_array('mix_ratio_array', self._mix_ratio_array)

        return gas_entry



from taurex.log import Logger
from taurex.constants import G, RJUP, MJUP, RSOL, MSOL
from taurex.data.fittable import fitparam, Fittable
import numpy as np
from taurex.util.emission import black_body
from taurex.output.writeable import Writeable


class Star(Fittable, Logger, Writeable):
    """
    A base class that holds information on the star in the model.
    Its implementation is a star that has a blackbody spectrum.


    Parameters
    ----------

    temperature: float, optional
        Stellar temperature in Kelvin

    radius: float, optional
        Stellar radius in Solar radius

    metallicity: float, optional
        Metallicity in solar values

    mass: float, optional
        Stellar mass in solar mass

    distance: float, optional
        Distance from Earth in pc

    magnitudeK: float, optional
        Maginitude in K band


    """

    def __init__(self, temperature=5000, radius=1.0, distance=1,
                 magnitudeK=10.0, mass=1.0, metallicity=1.0):

        Logger.__init__(self, self.__class__.__name__)
        Fittable.__init__(self)
        self._temperature = temperature
        self._radius = radius*RSOL
        self._mass = mass*MSOL
        self.debug('Star mass %s', self._mass)
        self.sed = None
        self.distance = distance
        self.magnitudeK = magnitudeK
        self._metallicity = metallicity

    @property
    def radius(self):
        """
        Radius in metres

        Returns
        -------
        R: float

        """
        return self._radius

    @property
    def temperature(self):
        """
        Blackbody temperature in Kelvin

        Returns
        -------
        T: float

        """
        return self._temperature

    @temperature.setter
    def temperature(self, value):
        self._temperature = value

    @property
    def mass(self):
        return self._mass

    def initialize(self, wngrid):
        """
        Initializes the blackbody spectrum on the given wavenumber grid

        Parameters
        ----------
        wngrid: :obj:`array`
            Wavenumber grid cm-1 to compute black body spectrum

        """
        self.sed = black_body(wngrid, self.temperature)

    @property
    def spectralEmissionDensity(self):
        """
        Spectral emmision density

        Returns
        -------
        sed: :obj:`array`
        """
        return self.sed

    def write(self, output):
        star = output.create_group('Star')
        star.write_string('star_type', self.__class__.__name__)
        star.write_scalar('temperature', self.temperature)
        star.write_scalar('radius', self._radius/RSOL)
        star.write_scalar('distance', self.distance)
        star.write_scalar('mass', self._mass/MSOL)
        star.write_scalar('magnitudeK', self.magnitudeK)
        star.write_scalar('metallicity', self._metallicity)
        star.write_scalar('radius_m', self.radius)
        star.write_array('SED', self.spectralEmissionDensity)
        star.write_scalar('mass_kg', self._mass)
        return star


class BlackbodyStar(Star):
    """Alias for the base star type"""
    pass



from .temparray import TemperatureArray
import numpy as np


class TemperatureFile(TemperatureArray):
    """
    A temperature profile read from file

    Parameters
    ----------

    filename : str
        File name for temperature profile

    """

    def __init__(self, filename=None):
        super().__init__(tp_array=np.loadtxt(filename))



from .optimizer import Optimizer
import pypolychord
from pypolychord.settings import PolyChordSettings
from pypolychord.priors import UniformPrior
import numpy as np
import os

from taurex.util.util import read_table, read_error_line, read_error_into_dict, quantile_corner, recursively_save_dict_contents_to_output, weighted_avg_and_std


class PolyChordOptimizer(Optimizer):

    def __init__(self, polychord_path=None, observed=None, model=None,
                 num_live_points=1500,
                 max_iterations=0,
                 maximum_modes=100,
                 cluster=True,
                 evidence_tolerance=0.5,
                 mode_tolerance=-1e90,
                 resume=False,
                 verbosity=1, sigma_fraction=0.1):
        super().__init__('Multinest', observed, model, sigma_fraction)

        # number of live points
        self.n_live_points = int(num_live_points)
        # maximum no. of iterations (0=inf)
        self.max_iter = int(max_iterations)
        # search for multiple modes
        # parameters on which to cluster, e.g. if nclust_par = 3, it will cluster on the first 3 parameters only.
        # if ncluster_par = -1 it clusters on all parameters
        # maximum number of modes
        self.max_modes = int(maximum_modes)
        self.do_clustering = cluster
        # run in constant efficiency mode
        # set log likelihood tolerance. If change is smaller, multinest will have converged
        self.evidence_tolerance = evidence_tolerance
        self.mode_tolerance = mode_tolerance
        # importance nested sampling

        self.dir_polychord = polychord_path
        self._polychord_output = None
        self.resume = resume
        self.verbose = verbosity

    def compute_fit(self):
        self._polychord_output = None
        data = self._observed.spectrum
        datastd = self._observed.errorBar
        sqrtpi = np.sqrt(2*np.pi)

        ndim = len(self.fitting_parameters)

        def polychord_loglike(cube):
            # log-likelihood function called by polychord
            fit_params_container = np.array(
                [cube[i] for i in range(len(self.fitting_parameters))])
            chi_t = self.chisq_trans(fit_params_container, data, datastd)

            # print('---------START---------')
            # print('chi_t',chi_t)
            # print('LOG',loglike)
            loglike = -np.sum(np.log(datastd*sqrtpi)) - 0.5 * chi_t
            return loglike, [0.0]

        def polychord_uniform_prior(hypercube):
            # prior distributions called by polychord. Implements a uniform prior
            # converting parameters from normalised grid to uniform prior
            # print(type(cube))
            cube = [0.0]*ndim

            for idx, bounds in enumerate(self.fit_boundaries):
                # print(idx,self.fitting_parameters[idx])
                bound_min, bound_max = bounds
                cube[idx] = (hypercube[idx] *
                             (bound_max-bound_min)) + bound_min
                #print('CUBE idx',cube[idx])
            # print('-----------')
            return cube
        status = None

        datastd_mean = np.mean(datastd)

        settings = PolyChordSettings(ndim, 1)
        settings.nlive = ndim * 25
        settings.num_repeats = ndim * 5
        settings.do_clustering = self.do_clustering
        settings.num_repeats = ndim
        settings.precision_criterion = self.evidence_tolerance
        settings.logzero = -1e70
        settings.read_resume = self.resume
        settings.base_dir = self.dir_polychord
        settings.file_root = '1-'
        self.warning('Number of dimensions {}'.format(ndim))
        self.warning('Fitting parameters {}'.format(self.fitting_parameters))

        self.info('Beginning fit......')
        pypolychord.run_polychord(
            polychord_loglike, ndim, 1, settings, polychord_uniform_prior)
        self._polychord_output = self.store_polychord_solutions()
        print(self._polychord_output)

    def write_optimizer(self, output):
        opt = super().write_optimizer(output)

        # sampling efficiency (parameter, ...)
        opt.write_scalar('do_clustering', int(self.do_clustering))
        # run in constant efficiency mode
        # set log likelihood tolerance. If change is smaller, multinest will have converged
        opt.write_scalar('evidence_tolerance', self.evidence_tolerance)
        opt.write_scalar('mode_tolerance', self.mode_tolerance)
        # importance nested samplin

        return opt

    def write_fit(self, output):
        fit = super().write_fit(output)

        # if self._polychord_output:
        #     recursively_save_dict_contents_to_output(output,self._polychord_output)

        return fit

    def store_polychord_solutions(self):

        self.warning('Store the polychord results')
        NEST_out = {'solutions': {}}
        data = np.loadtxt(os.path.join(self.dir_polychord, '1-.txt'))

        # self.get_poly_cluster_number(self.dir_polychord)
        NEST_stats = self.get_poly_stats(self.dir_polychord)
        NEST_out['NEST_POLY_stats'] = NEST_stats
        NEST_out['global_logE'] = (NEST_out['NEST_POLY_stats']['global evidence'],
                                   NEST_out['NEST_POLY_stats']['global evidence error'])

        modes_array = []
        modes_weights = []
        num_fit_params = len(self.fit_names)
        num_clusters = 1

        if self.do_clustering:
            # if clustering is switched on, checking how many clusters exist
            num_clusters = self.get_poly_cluster_number(self.dir_polychord)
            print('NUM_CLUSTERS: ', num_clusters)
            if num_clusters == 1:
                # Get chains directly from file 1-.txt
                modes_array = [data[:, 2:num_fit_params+2]]
                modes_weights = [data[:, 0]]
            else:
                # cycle through clusters
                for midx in range(num_clusters):
                    data = np.loadtxt(os.path.join(
                        self.dir_polychord, 'clusters/1-_{0}.txt'.format(midx+1)))
                    modes_array.append(data[:, 2:num_fit_params+2])
                    modes_weights.append(data[:, 0])
        else:
            # Get chains directly from file 1-.txt
            modes_array = [data[:, 2:num_fit_params+2]]
            modes_weights = [data[:, 0]]

        modes_array = np.asarray(modes_array)
        modes_weights = np.asarray(modes_weights)

        for nmode in range(num_clusters):

            mydict = {'type': 'nest_poly',
                      'local_logE': (NEST_out['NEST_POLY_stats']['modes'][0]['local log-evidence'],  NEST_out['NEST_POLY_stats']['modes'][0]['local log-evidence error']),
                      'weights': np.asarray(modes_weights[nmode]),
                      'tracedata': modes_array[nmode],
                      'fit_params': {}}

            for idx, param_name in enumerate(self.fit_names):

                trace = modes_array[nmode][:, idx]
                q_16, q_50, q_84 = quantile_corner(trace, [0.16, 0.5, 0.84],
                                                   weights=np.asarray(modes_weights[nmode]))
                mydict['fit_params'][param_name] = {
                    'value': q_50,
                    'sigma_m': q_50-q_16,
                    'sigma_p': q_84-q_50,
                    'nest_map': NEST_stats['modes'][nmode]['maximum a posterior'][idx],
                    'nest_mean': NEST_stats['modes'][nmode]['mean'][idx],
                    'nest_sigma': NEST_stats['modes'][nmode]['sigma'][idx],
                    'trace': trace,
                }

            NEST_out['solutions']['solution{}'.format(nmode)] = mydict

        return NEST_out

    def get_poly_cluster_number(self, dir):
        import glob
        '''counts polychord cluster files in 'clusters' folder'''
        cluster_list = glob.glob(os.path.join(dir, 'clusters/1-*.txt'))
        c_idx = []
        for file in cluster_list:
            if file[-5].isdigit():
                c_idx.append(int(file[-5]))
        try:
            num = np.max(c_idx)
        except ValueError:
            num = 1
        return num

    def get_poly_stats(self, dir):
        '''replicates some of PyMultiNest.Analyzer for PolyChord'''
        stats = {}
        stats['modes'] = {}

        # re-count number of cluster files
        if self.do_clustering:
            num_clusters = self.get_poly_cluster_number(dir)
        else:
            num_clusters = 1

        nmode = 0  # mode index
        # open .stats file and reading global/local evidences
        with open(os.path.join(self.dir_polychord, '1-.stats')) as f:
            lines = f.readlines()
            for idx, line in enumerate(lines):
                if idx == 8:  # skip to global evidence line
                    tmp_line = line.split()
                    gL_mu = tmp_line[2]
                    gL_sig = tmp_line[-1]
                    stats['global evidence'] = gL_mu
                    stats['global evidence error'] = gL_sig
                if idx > 13:  # skip to local evidence
                    tmp_line = line.split()
                    stats['modes'][nmode] = {}
                    stats['modes'][nmode]['local log-evidence'] = tmp_line[3]
                    stats['modes'][nmode]['local log-evidence error'] = tmp_line[5]
                    nmode += 1
                    # stopping reading file when clusters are exhausted
                    if idx == (13 + num_clusters):
                        break

        # opening cluster files (or global file if no clustering) and get MAP, mean, sigma for parameters
        if self.do_clustering:
            for midx in range(num_clusters):
                # cycling through cluster files
                data = np.loadtxt(os.path.join(
                    self.dir_polychord, 'clusters/1-_{0}.txt'.format(midx+1)))
                # find maximum likelihood index
                mL_idx = np.where(data[:, 1] == np.min(data[:, 1]))
                stats['modes'][midx]['maximum a posterior'] = {}
                stats['modes'][midx]['mean'] = {}
                stats['modes'][midx]['sigma'] = {}
                for idx in range(len(self.fit_names)):
                    # cycle through parameters
                    # maximum likelihood values
                    stats['modes'][midx]['maximum a posterior'][idx] = data[mL_idx, 2+idx]
                    # weighted average and sigma
                    mu, sig = weighted_avg_and_std(data[:, 2+idx], data[:, 0])
                    stats['modes'][midx]['mean'][idx] = mu
                    stats['modes'][midx]['sigma'][idx] = sig

        return stats

    def sample_parameters(self, solution):
        from taurex.util.util import random_int_iter
        solution_id = 'solution{}'.format(solution)
        samples = self._polychord_output['solutions'][solution_id]['tracedata']
        weights = self._polychord_output['solutions'][solution_id]['weights']

        for x in random_int_iter(samples.shape[0], self._sigma_fraction):
            w = weights[x]+1e-300

            yield samples[x, :], w

    def get_solution(self):
        names = self.fit_names
        opt_values = self.fit_values
        opt_map = self.fit_values
        solutions = [
            (k, v) for k, v in self._polychord_output['solutions'].items() if 'solution' in k]

        for k, v in solutions:
            solution_idx = int(k[8:])
            for p_name, p_value in v['fit_params'].items():
                if p_name in ('mu_derived',):
                    continue
                idx = names.index(p_name)
                opt_map[idx] = p_value['nest_map']
                opt_values[idx] = p_value['value']

            yield solution_idx, opt_map, opt_values, [
                ('fit_params', v['fit_params']),
                ('tracedata', v['tracedata']),
                ('weights', v['weights'])]



"""Includes some useful definitions and enums"""
from enum import IntEnum


class OutputSize(IntEnum):
    """
    This enum describes how heavy the final output will be
    """

    heavy = 6
    """Output everything native and binned (large file size)"""
    light = 3
    """Output everything but only keep binned optical depths"""
    lighter = 1
    """Do not output optical depths"""


"""
Modules dealing with reading data from observations
"""

from .spectrum import BaseSpectrum
from .observed import ObservedSpectrum
from .array import ArraySpectrum
from .taurex import TaurexSpectrum



"""
Base contribution classes and functions for computing optical depth
"""

from taurex.log import Logger
from taurex.data.fittable import Fittable
import numpy as np
from taurex.output.writeable import Writeable
import numba


@numba.jit(nopython=True, nogil=True)
def contribute_tau(startK, endK, density_offset, sigma, density, path, nlayers,
                   ngrid, layer, tau):
    """
    Generic cross-section integration function for tau, numba-fied for
    performance.

    This has the form:

    .. math::

        \\tau_{\\lambda}(z) = \\int_{z_{0}}^{z_{1}} \\sigma(z') \\rho(z') dz',

    where :math:`z` is the layer, :math:`z_0` and :math:`z_1` are ``startK``
    and ``endK`` respectively. :math:`\\sigma` is the weighted
    cross-section ``sigma``. :math:`rho` is the ``density`` and
    :math:`dz'` is the integration path length ``path``


    Parameters
    ----------
    startK: int
        starting layer in integration

    endK: int
        last layer in integration

    density_offset: int
        Which part of the density profile to start from

    sigma: :obj:`array`
        cross-section

    density: array_like
        density profile of atmosphere

    path: array_like
        path-length or altitude gradient

    nlayers: int
        Total number of layers (unused)

    ngrid: int
        total number of grid points

    layer: int
        Which layer we currently on

    Returns
    -------
    tau : array_like
        optical depth (well almost you still need to do ``exp(-tau)`` yourself)

    """
    for k in range(startK, endK):
        _path = path[k]
        _density = density[k+density_offset]
        # for mol in range(nmols):
        for wn in range(ngrid):
            tau[layer, wn] += sigma[k+layer, wn]*_path*_density


class Contribution(Fittable, Logger, Writeable):
    """

    *Abstract class*

    The base class for modelling contributions to the optical depth.
    By default this handles contributions from cross-sections.
    If the type of contribution being implemented is a `sigma`-type
    like the form given in :func:`contribute_tau` then
    To function in Taurex3, it only requires the concrete implementation of:

    - :func:`prepare_each`

    Different forms may require reimplementing
    :func:`contribute` as well as :func:`prepare`

    Parameters
    ----------
    name : str
        Identifier of the contribution.

    """

    def __init__(self, name):
        Logger.__init__(self, name)
        Fittable.__init__(self)
        self._name = name
        self._total_contribution = None
        self._enabled = True
        self.sigma_xsec = None

    @property
    def order(self):
        """
        Computational order. Lower numbers are given
        higher priority and are computed first.

        Returns
        -------
        int:
            Order of computation

        """
        return 5

    @property
    def name(self):
        """Name of the contribution. Identifier for plots"""
        return self._name

    def contribute(self, model, start_layer, end_layer,
                   density_offset, layer, density, tau, path_length=None):
        """
        Computes an integral for a single layer for the optical depth.

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            A forward model

        start_layer: int
            Lowest layer limit for integration

        end_layer: int
            Upper layer limit of integration

        density_offset: int
            offset in density layer

        layer: int
            atmospheric layer being computed

        density: :obj:`array`
            density profile of atmosphere

        tau: :obj:`array`
            optical depth to store result

        path_length: :obj:`array`
            integration length

        """
        self.debug('SIGMA %s', self.sigma_xsec.shape)
        self.debug(' %s %s %s %s %s %s %s', start_layer, end_layer,
                   density_offset, layer, density, tau, self._ngrid)
        contribute_tau(start_layer, end_layer, density_offset,
                       self.sigma_xsec, density, path_length, self._nlayers,
                       self._ngrid, layer, tau)
        self.debug('DONE')

    def build(self, model):
        """
        Called during forward model build phase
        Does nothing by default

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        """
        pass

    def prepare_each(self, model, wngrid):
        """
        **Requires implementation**

        Used to prepare each component of the contribution.
        For context when the main ``taurex`` program is run
        with the option each spectra is the component for the
        contribution. For cross-section based contributions,
        the components are each molecule
        Should yield the name of the component and the component itself

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            Name of component and component itself

        """

        raise NotImplementedError

    def prepare(self, model, wngrid):
        """

        Used to prepare the contribution for the calculation.
        Called before the forward model performs the main optical depth
        calculation. Default behaviour is to loop through :func:`prepare_each`
        and sum all results into a single cross-section.

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid
        """

        self._ngrid = wngrid.shape[0]
        self._nlayers = model.nLayers

        sigma_xsec = np.zeros(shape=(self._nlayers, self._ngrid))

        for gas, sigma in self.prepare_each(model, wngrid):
            self.debug('Gas %s', gas)
            self.debug('Sigma %s', sigma)
            sigma_xsec += sigma

        self.sigma_xsec = sigma_xsec
        self.debug('Final sigma is %s', self.sigma_xsec)
        self.info('Done')

    def finalize(self, model, tau):
        """
        Called in the last phase of the calculation, after the optical
        depth has be completely computed.
        """
        pass

    @property
    def sigma(self):
        return self.sigma_xsec

    def write(self, output):
        """
        Writes contribution class and arguments to file

        Parameters
        ----------
        output: :class:`~taurex.output.output.Output`
        """
        contrib = output.create_group(self.__class__.__name__)
        return contrib



from .polychord import PolyChordOptimizer
import numpy as np
import os
# Import some example python likelihoods
import dyPolyChord.python_likelihoods as likelihoods
import dyPolyChord.python_priors as priors  # Import some example python priors
import dyPolyChord.pypolychord_utils
import dyPolyChord
import time


class dyPolyChordOptimizer(PolyChordOptimizer):

    def __init__(self, polychord_path=None, observed=None, model=None,
                 num_live_points=1500,
                 max_iterations=0,
                 maximum_modes=100,
                 cluster=True,
                 evidence_tolerance=0.5,
                 mode_tolerance=-1e90,
                 resume=False,
                 verbosity=1, sigma_fraction=0.1):
        super().__init__(
            polychord_path=polychord_path, observed=observed, model=model,
            num_live_points=num_live_points,
            max_iterations=max_iterations,
            maximum_modes=maximum_modes,
            cluster=cluster,
            evidence_tolerance=evidence_tolerance,
            mode_tolerance=mode_tolerance,
            resume=resume,
            verbosity=verbosity, sigma_fraction=sigma_fraction)

    def compute_fit(self):
        self._polychord_output = None
        data = self._observed.spectrum
        datastd = self._observed.errorBar
        sqrtpi = np.sqrt(2*np.pi)

        ndim = len(self.fitting_parameters)

        def polychord_loglike(cube):
            # log-likelihood function called by polychord
            fit_params_container = np.array(
                [cube[i] for i in range(len(self.fitting_parameters))])
            chi_t = self.chisq_trans(fit_params_container, data, datastd)

            # print('---------START---------')
            # print('chi_t',chi_t)
            # print('LOG',loglike)
            loglike = -np.sum(np.log(datastd*sqrtpi)) - 0.5 * chi_t
            return loglike, [0.0]

        def polychord_uniform_prior(hypercube):
            # prior distributions called by polychord. Implements a uniform prior
            # converting parameters from normalised grid to uniform prior
            # print(type(cube))
            cube = [0.0]*ndim

            for idx, bounds in enumerate(self.fit_boundaries):
                # print(idx,self.fitting_parameters[idx])
                bound_min, bound_max = bounds
                cube[idx] = (hypercube[idx] *
                             (bound_max-bound_min)) + bound_min
                #print('CUBE idx',cube[idx])
            # print('-----------')
            return cube
        status = None

        datastd_mean = np.mean(datastd)

        likelihood = polychord_loglike
        prior = polychord_uniform_prior

        # Make a callable for running PolyChord
        my_callable = dyPolyChord.pypolychord_utils.RunPyPolyChord(
            likelihood, prior, ndim)

        # Specify sampler settings (see run_dynamic_ns.py documentation for more details)
        # whether to maximise parameter estimation or evidence accuracy.
        dynamic_goal = 1.0
        # number of live points to use in initial exploratory run.
        ninit = ndim*5
        # total computational budget is the same as standard nested sampling with nlive_const live points.
        nlive_const = ndim * 25
        settings_dict = {
            'num_repeats': ndim * 5,
            'do_clustering': self.do_clustering,
            'num_repeats': ndim,
            'precision_criterion': self.evidence_tolerance,
            'logzero': -1e70,
            'read_resume': self.resume,
            'base_dir': self.dir_polychord,
            'file_root': '1-'}

        try:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD

            # Run dyPolyChord
            dyPolyChord.run_dypolychord(my_callable, dynamic_goal, settings_dict,
                                        ninit=ninit, nlive_const=nlive_const, comm=comm)
        except ImportError:
            dyPolyChord.run_dypolychord(my_callable, dynamic_goal, settings_dict,
                                        ninit=ninit, nlive_const=nlive_const)

        time.sleep(2.0)

        #pypolychord.run_polychord(polychord_loglike, ndim, 1, settings, polychord_uniform_prior)
        self._polychord_output = self.store_polychord_solutions()



from .contribution import Contribution
import numpy as np
from taurex.data.fittable import fitparam


class LeeMieContribution(Contribution):
    """
    Computes Mie scattering contribution to optica depth
    Formalism taken from: Lee et al. 2013, ApJ, 778, 97

    Parameters
    ----------

    lee_mie_radius: float
        Particle radius in um

    lee_mie_q: float
        Extinction coefficient

    lee_mie_mix_ratio: float
        Mixing ratio in atmosphere

    lee_mie_bottomP: float
        Bottom of cloud deck in Pa

    lee_mie_topP: float
        Top of cloud deck in Pa


    """

    def __init__(self, lee_mie_radius=0.01, lee_mie_q=40,
                 lee_mie_mix_ratio=1e-10, lee_mie_bottomP=-1,
                 lee_mie_topP=-1):
        super().__init__('Mie')

        self._mie_radius = lee_mie_radius
        self._mie_q = lee_mie_q
        self._mie_mix = lee_mie_mix_ratio
        self._mie_bottom_pressure = lee_mie_bottomP
        self._mie_top_pressure = lee_mie_topP

    @fitparam(param_name='lee_mie_radius',
              param_latex='$R^{lee}_{\mathrm{mie}}$',
              default_fit=False,
              default_bounds=[0.01, 0.5])
    def mieRadius(self):
        """
        Particle radius in um
        """
        return self._mie_radius

    @mieRadius.setter
    def mieRadius(self, value):
        self._mie_radius = value

    @fitparam(param_name='lee_mie_q', param_latex='$Q_\mathrm{ext}$',
              default_fit=False, default_bounds=[-10, 1])
    def mieQ(self):
        """
        Extinction coefficient
        """
        return self._mie_q

    @mieQ.setter
    def mieQ(self, value):
        self._mie_q = value

    @fitparam(param_name='lee_mie_topP',
              param_latex='$P^{lee}_\mathrm{top}$',
              default_mode='log',
              default_fit=False,
              default_bounds=[-1, 1])
    def mieTopPressure(self):
        """
        Pressure at top of cloud deck in Pa
        """
        return self._mie_top_pressure

    @mieTopPressure.setter
    def mieTopPressure(self, value):
        self._mie_top_pressure = value

    @fitparam(param_name='lee_mie_bottomP',
              param_latex='$P^{lee}_\mathrm{bottom}$',
              default_mode='log',
              default_fit=False,
              default_bounds=[-1, 1])
    def mieBottomPressure(self):
        """
        Pressure at bottom of cloud deck in Pa
        """
        return self._mie_bottom_pressure

    @mieBottomPressure.setter
    def mieBottomPressure(self, value):
        self._mie_bottom_pressure = value

    @fitparam(param_name='lee_mie_mix_ratio',
              param_latex='$\chi^{lee}_\mathrm{mie}$',
              default_mode='log',
              default_fit=False,
              default_bounds=[-1, 1])
    def mieMixing(self):
        """
        Mixing ratio in atmosphere
        """
        return self._mie_mix

    @mieMixing.setter
    def mieMixing(self, value):
        self._mie_mix = value

    def prepare_each(self, model, wngrid):
        """
        Computes and weights the mie opacity for
        the pressure regions given

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            ``Lee`` and the weighted mie opacity.

        """
        self._nlayers = model.nLayers
        self._ngrid = wngrid.shape[0]

        pressure_profile = model.pressureProfile

        bottom_pressure = self.mieBottomPressure
        if bottom_pressure < 0:
            bottom_pressure = pressure_profile[0]

        top_pressure = self.mieTopPressure
        if top_pressure < 0:
            top_pressure = pressure_profile[-1]

        wltmp = 10000/wngrid

        a = self.mieRadius

        x = 2.0 * np.pi * a / wltmp
        self.debug('wngrid %s', wngrid)
        self.debug('x %s', x)
        Qext = 5.0 / (self.mieQ * x**(-4.0) + x**(0.2))

        sigma_xsec = np.zeros(shape=(self._nlayers, wngrid.shape[0]))

        # This must transform um to the xsec format in TauREx (m2)
        am = a * 1e-6

        sigma_mie = Qext * np.pi * (am**2.0)

        self.debug('Qext %s', Qext)
        self.debug('radius um %s', a)
        self.debug('sigma %s', sigma_mie)

        self.debug('bottome_pressure %s', bottom_pressure)
        self.debug('top_pressure %s', top_pressure)

        cloud_filter = (pressure_profile <= bottom_pressure) & \
            (pressure_profile >= top_pressure)

        sigma_xsec[cloud_filter, ...] = sigma_mie * self.mieMixing

        self.sigma_xsec = sigma_xsec

        self.debug('final xsec %s', self.sigma_xsec)

        yield 'Lee', sigma_xsec

    def write(self, output):
        contrib = super().write(output)
        contrib.write_scalar('lee_mie_radius', self._mie_radius)
        contrib.write_scalar('lee_mie_q', self._mie_q)
        contrib.write_scalar('lee_mie_mix_ratio', self._mie_mix)
        contrib.write_scalar('lee_mie_bottomP', self._mie_bottom_pressure)
        contrib.write_scalar('lee_mie_topP', self._mie_top_pressure)
        return contrib



from .array import ArraySpectrum
import numpy as np
from taurex.util.util import wnwidth_to_wlwidth


class TaurexSpectrum(ArraySpectrum):
    """
    Observation is a taurex spectrum from a HDF5 file

    An instrument function must have been used for this to work

    Parameters
    -----------
    filename: string
        Path to taurex spectrum HDF5 output.

    """

    def __init__(self, filename):
        super().__init__(self._load_from_hdf5(filename))

    def _load_from_hdf5(self, filename):
        import h5py

        with h5py.File(filename, 'r') as f:
            try:
                wngrid = f['Output']['Spectra']['instrument_wngrid'][:]
            except KeyError:
                self.error('Could not find instrument outputs in HDF5, '
                           'this was caused either by the HDF5 being a '
                           'retrieval output or not running with some '
                           'form of instrument in the forward model'
                           ' input par file')
                raise KeyError('Instrument output not found')

            spectrum = f['Output']['Spectra']['instrument_spectrum'][:]
            noise = f['Output']['Spectra']['instrument_noise'][:]
            wnwidth = f['Output']['Spectra']['instrument_wnwidth'][:]

        wlgrid = 10000/wngrid

        wlwidth = wnwidth_to_wlwidth(wngrid, wnwidth)

        return np.vstack((wlgrid, spectrum, noise, wlwidth)).T



from taurex.log import Logger
from taurex.data.fittable import Fittable
from taurex.output.writeable import Writeable


class ForwardModel(Logger, Fittable, Writeable):
    """A base class for producing forward models"""

    def __init__(self, name):
        Logger.__init__(self, name)
        Fittable.__init__(self)
        self.opacity_dict = {}
        self.cia_dict = {}

        self._native_grid = None

        self._fitting_parameters = {}

        self.contribution_list = []

    def __getitem__(self, key):
        return self._fitting_parameters[key][2]()

    def __setitem__(self, key, value):
        return self._fitting_parameters[key][3](value)

    def defaultBinner(self):
        from taurex.binning import NativeBinner
        return NativeBinner()

    def add_contribution(self, contrib):
        from taurex.contributions import Contribution
        if not isinstance(contrib, Contribution):
            raise TypeError('Is not a a contribution type')
        else:
            if not contrib in self.contribution_list:
                self.contribution_list.append(contrib)
            else:
                raise Exception('Contribution already exists')

    def build(self):
        raise NotImplementedError

    def model(self, wngrid=None, cutoff_grid=True):
        """Computes the forward model for a wngrid"""
        raise NotImplementedError

    def model_full_contrib(self, wngrid=None, cutoff_grid=True):
        """Computes the forward model for a wngrid for each contribution"""
        raise NotImplementedError

    @property
    def fittingParameters(self):
        return self._fitting_parameters

    def compute_error(self,  samples, wngrid=None, binner=None):
        return {}, {}

    def write(self, output):
        model = output.create_group('ModelParameters')
        model.write_string('model_type', self.__class__.__name__)
        contrib = model.create_group('Contributions')
        for c in self.contribution_list:
            c.write(contrib)

        return model



from .contribution import Contribution
import numpy as np
from taurex.data.fittable import fitparam


class FlatMieContribution(Contribution):
    """
    Computes a flat absorption contribution
    across all wavelengths to the optical depth

    Parameters
    ----------

    flat_mix_ratio: float
        Opacity value

    flat_bottomP: float
        Bottom of absorbing region in Pa

    flat_topP: float
        Top of absorbing region in Pa

    """

    def __init__(self,
                 flat_mix_ratio=1e-10, flat_bottomP=-1,
                 flat_topP=-1):
        super().__init__('Mie')

        self._mie_mix = flat_mix_ratio
        self._mie_bottom_pressure = flat_bottomP
        self._mie_top_pressure = flat_topP

    @fitparam(param_name='flat_topP',
              param_latex='$P^{mie}_\mathrm{top}$',
              default_mode='log',
              default_fit=False,
              default_bounds=[1e-20, 1])
    def mieTopPressure(self):
        """
        Pressure at top of absorbing region in Pa
        """
        return self._mie_top_pressure

    @mieTopPressure.setter
    def mieTopPressure(self, value):
        self._mie_top_pressure = value

    @fitparam(param_name='flat_bottomP',
              param_latex='$P^{mie}_\mathrm{bottom}$',
              default_mode='log',
              default_fit=False,
              default_bounds=[1e-20, 1])
    def mieBottomPressure(self):
        """
        Pressure at bottom of absorbing region in Pa
        """
        return self._mie_bottom_pressure

    @mieBottomPressure.setter
    def mieBottomPressure(self, value):
        self._mie_bottom_pressure = value

    @fitparam(param_name='flat_mix_ratio',
              param_latex='$\chi_\mathrm{mie}$',
              default_mode='log',
              default_fit=False,
              default_bounds=[1e-20, 1])
    def mieMixing(self):
        """
        Opacity of absorbing region in m2
        """
        return self._mie_mix

    @mieMixing.setter
    def mieMixing(self, value):
        self._mie_mix = value

    def prepare_each(self, model, wngrid):
        """
        Computes and flat absorbing opacity for
        the pressure regions given

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            ``Flat`` and the weighted mie opacity.


        """
        self._nlayers = model.nLayers
        self._ngrid = wngrid.shape[0]

        pressure_profile = model.pressureProfile

        bottom_pressure = self.mieBottomPressure
        if bottom_pressure < 0:
            bottom_pressure = pressure_profile[0]

        top_pressure = self.mieTopPressure
        if top_pressure < 0:
            top_pressure = pressure_profile[-1]

        sigma_xsec = np.zeros(shape=(self._nlayers, wngrid.shape[0]))

        cloud_filter = (pressure_profile <= bottom_pressure) & \
                       (pressure_profile >= top_pressure)

        sigma_xsec[cloud_filter, ...] = \
            np.ones(shape=wngrid.shape) * self.mieMixing

        self.sigma_xsec = sigma_xsec

        yield 'Flat', sigma_xsec

    def write(self, output):
        contrib = super().write(output)
        contrib.write_scalar('flat_mix_ratio', self._mie_mix)
        contrib.write_scalar('flat_bottomP', self._mie_bottom_pressure)
        contrib.write_scalar('flat_topP', self._mie_top_pressure)
        return contrib



from .binner import Binner
from taurex import OutputSize


class LightcurveBinner(Binner):
    """
    A special class of binning used to generate the correct spectrum output.
    This is essentially the same as 
    :class:`~taurex.binning.nativebinner.NativeBinner`
    but for lightcurve forward models.


    """

    def bindown(self, wngrid, spectrum, grid_width=None, error=None):
        """Does nothing, only returns function arguments"""
        return wngrid, spectrum, error, grid_width

    def generate_spectrum_output(self, model_output,
                                 output_size=OutputSize.heavy):
        """
        Accepts only a lightcurve forward model. Stores the lightcurve
        as well as the spectrum.

        Parameters
        ----------
        model_output: obj:`tuple`
            Result from running a TauREx3 lightcurve forward model

        output_size: :class:`~taurex.taurexdefs.OutputSize`
            Size of the output.


        Returns
        -------
        :obj:`dict`:
            Dictionary of spectra containing both lightcurves
            and spectra.


        """

        output = {}

        wngrid, lightcurve, tau, extra = model_output
        native_grid, native, binned, extra = extra

        output['native_wngrid'] = native_grid
        output['native_wlgrid'] = 10000/native_grid
        output['binned_wngrid'] = wngrid
        output['binned_wlgrid'] = 10000/wngrid
        output['lightcurve'] = lightcurve
        output['native_spectrum'] = native
        output['binned_spectrum'] = binned
        if output_size > OutputSize.lighter:
            output['binned_tau'] = self.bindown(wngrid, tau)[1]
            if output_size > OutputSize.light:
                output['native_tau'] = tau

        return output



from .interpolateopacity import InterpolatingOpacity
import pickle
import numpy as np
import pathlib


class HDF5Opacity(InterpolatingOpacity):
    """
    This is the base class for computing opactities

    """

    def __init__(self, filename, interpolation_mode='exp', in_memory=False):
        super().__init__('HDF5Opacity:{}'.format(pathlib.Path(filename).stem[0:10]),
                         interpolation_mode=interpolation_mode)

        self._filename = filename
        self._molecule_name = None
        self._spec_dict = None
        self.in_memory = in_memory
        self._load_hdf_file(filename)

    @property
    def moleculeName(self):
        return self._molecule_name

    @property
    def xsecGrid(self):
        return self._xsec_grid

    def _load_hdf_file(self, filename):
        import h5py
        import astropy.units as u
        # Load the pickle file
        self.debug('Loading opacity from {}'.format(filename))

        self._spec_dict = h5py.File(filename, 'r')

        self._wavenumber_grid = self._spec_dict['bin_edges'][:]

        #temperature_units = self._spec_dict['t'].attrs['units']
        #t_conversion = u.Unit(temperature_units).to(u.K).value

        self._temperature_grid = self._spec_dict['t'][:]  # *t_conversion

        pressure_units = self._spec_dict['p'].attrs['units']
        try:
            p_conversion = u.Unit(pressure_units).to(u.Pa)
        except:
            p_conversion = u.Unit(pressure_units, format="cds").to(u.Pa)

        self._pressure_grid = self._spec_dict['p'][:]*p_conversion

        if self.in_memory:
            self._xsec_grid = self._spec_dict['xsecarr'][...]
        else:
            self._xsec_grid = self._spec_dict['xsecarr']

        self._resolution = np.average(np.diff(self._wavenumber_grid))
        self._molecule_name = self._spec_dict['mol_name'][()]

        if isinstance(self._molecule_name, np.ndarray):
            self._molecule_name = self._molecule_name[0]

        self._min_pressure = self._pressure_grid.min()
        self._max_pressure = self._pressure_grid.max()
        self._min_temperature = self._temperature_grid.min()
        self._max_temperature = self._temperature_grid.max()

    @property
    def wavenumberGrid(self):
        return self._wavenumber_grid

    @property
    def temperatureGrid(self):
        return self._temperature_grid

    @property
    def pressureGrid(self):
        return self._pressure_grid

    @property
    def resolution(self):
        return self._resolution

        # return factor*(q_11*(Pmax-P)*(Tmax-T) + q_21*(P-Pmin)*(Tmax-T) + q_12*(Pmax-P)*(T-Tmin) + q_22*(P-Pmin)*(T-Tmin))



"""The main taurex program"""


def main():
    import argparse
    import datetime

    import logging
    from taurex.mpi import get_rank
    from taurex.log import setLogLevel
    from taurex.log.logger import root_logger
    from taurex.parameter import ParameterParser
    from taurex.output.hdf5 import HDF5Output
    from taurex.util.output import generate_profile_dict, store_contributions
    from .taurexdefs import OutputSize
    from . import __version__ as version

    import numpy as np

    parser = argparse.ArgumentParser(description='TauREx {}'.format(version))

    parser.add_argument("-i", "--input", dest='input_file', type=str,
                        required=True, help="Input par file to pass")

    parser.add_argument("-R", "--retrieval", dest='retrieval', default=False,
                        help="When set, runs retrieval", action='store_true')

    parser.add_argument("-p", "--plot", dest='plot', default=False,
                              help="Whether to plot after the run",
                              action='store_true')

    parser.add_argument("-g", "--debug-log", dest='debug', default=False,
                        help="Debug log output", action='store_true')

    parser.add_argument("-c", "--show-contrib", dest='contrib',
                        default=False, help="Show basic contributions",
                        action='store_true')

    parser.add_argument("-C", "--full-contrib", dest='full_contrib',
                        default=False, help="Show ALL contributions",
                        action='store_true')

    parser.add_argument("--light", dest='light', default=False,
                        help="Light outputs", action='store_true')

    parser.add_argument("--lighter", dest='lighter', default=False,
                        help="Even Lighter outputs", action='store_true')

    parser.add_argument("-o", "--output_file", dest='output_file', type=str)

    parser.add_argument("-S", "--save-spectrum",
                        dest='save_spectrum', type=str)
    args = parser.parse_args()

    output_size = OutputSize.heavy

    if args.light:
        output_size = OutputSize.light

    if args.lighter:
        output_size = OutputSize.lighter

    if args.debug:
        setLogLevel(logging.DEBUG)

    root_logger.info('TauREx %s', version)

    root_logger.info('TauREx PROGRAM START AT %s', datetime.datetime.now())

    # Parse the input file
    pp = ParameterParser()
    pp.read(args.input_file)

    # Setup global parameters
    pp.setup_globals()
    # Generate a model from the input
    model = pp.generate_appropriate_model()

    # build the model
    model.build()

    # Get the spectrum
    observation = pp.generate_observation()

    binning = pp.generate_binning()

    wngrid = None

    if binning == 'observed' and observation is None:
        logging.critical('Binning selected from Observation yet None provided')
        quit()

    if binning is None:
        if observation is None or observation == 'self':
            binning = model.defaultBinner()
            wngrid = model.nativeWavenumberGrid
        else:
            binning = observation.create_binner()
            wngrid = observation.wavenumberGrid
    else:
        if binning == 'native':
            binning = model.defaultBinner()
            wngrid = model.nativeWavenumberGrid
        elif binning == 'observed':
            binning = observation.create_binner()
            wngrid = observation.wavenumberGrid
        else:
            binning, wngrid = binning

    instrument = pp.generate_instrument(binner=binning)

    num_obs = 1
    if instrument is not None:
        instrument, num_obs = instrument

    if observation == 'self' and instrument is None:
        logging.getLogger('taurex').critical(
            'Instrument nust be specified when using self option')
        raise ValueError('No instruemnt specified for self option')

    inst_result = None
    if instrument is not None:
        inst_result = instrument.model_noise(
            model, model_res=model.model(), num_observations=num_obs)

    # Observation on self
    if observation == 'self':
        from taurex.data.spectrum import ArraySpectrum
        from taurex.util.util import wnwidth_to_wlwidth
        inst_wngrid, inst_spectrum, inst_noise, inst_width = inst_result

        inst_wlgrid = 10000/inst_wngrid

        inst_wlwidth = wnwidth_to_wlwidth(inst_wngrid, inst_width)
        observation = ArraySpectrum(
            np.vstack([inst_wlgrid, inst_spectrum,
                       inst_noise, inst_wlwidth]).T)
        binning = observation.create_binner()

    # Handle outputs
    if args.output_file:
        # Output taurex data
        with HDF5Output(args.output_file) as o:
            model.write(o)

    optimizer = None
    solution = None

    if args.retrieval is True:
        import time
        if observation is None:
            logging.critical('No spectrum is defined!!')
            quit()

        optimizer = pp.generate_optimizer()
        optimizer.set_model(model)
        optimizer.set_observed(observation)

        fitting_parameters = pp.generate_fitting_parameters()

        for key, value in fitting_parameters.items():
            fit = value['fit']
            bounds = value['bounds']
            mode = value['mode']
            factor = value['factor']

            if fit:
                logging.info('Fitting: {}'.format(key))
                optimizer.enable_fit(key)
            else:
                optimizer.disable_fit(key)

            if factor:
                optimizer.set_factor_boundary(key, factor)

            if bounds:
                optimizer.set_boundary(key, bounds)

            if mode:
                optimizer.set_mode(key, mode.lower())

        start_time = time.time()
        solution = optimizer.fit(output_size=output_size)

        end_time = time.time()

        root_logger.info(
            'Total Retrieval finish in %s seconds', end_time-start_time)

        for _, optimized, _, _ in optimizer.get_solution():
            optimizer.update_model(optimized)
            break

    result = model.model()

    if args.save_spectrum is not None:

        # with open(args.save_spectrum, 'w') as f:
        from taurex.util.util import wnwidth_to_wlwidth, compute_bin_edges
        save_wnwidth = compute_bin_edges(wngrid)[1]
        save_wl = 10000/wngrid
        save_wlwidth = wnwidth_to_wlwidth(wngrid, save_wnwidth)
        save_model = binning.bin_model(result)[1]
        save_error = np.zeros_like(save_wl)
        if inst_result is not None:
            inst_wngrid, inst_spectrum, inst_noise, inst_width = inst_result

            save_model = inst_spectrum
            save_wl = 10000/inst_wngrid

            save_wlwidth = wnwidth_to_wlwidth(inst_wngrid, inst_width)

            save_error = inst_noise

        np.savetxt(args.save_spectrum,
                   np.vstack((save_wl, save_model, save_error,
                              save_wlwidth)).T)

    if args.output_file:

        # Output taurex data
        with HDF5Output(args.output_file, append=True) as o:

            out = o.create_group('Output')
            if observation is not None:
                obs = o.create_group('Observed')
                observation.write(obs)

            profiles = generate_profile_dict(model)
            spectrum = \
                binning.generate_spectrum_output(result,
                                                 output_size=output_size)

            if inst_result is not None:
                spectrum['instrument_wngrid'] = inst_result[0]
                spectrum['instrument_wnwidth'] = inst_result[-1]
                spectrum['instrument_wlgrid'] = 10000/inst_result[0]
                spectrum['instrument_spectrum'] = inst_result[1]
                spectrum['instrument_noise'] = inst_result[2]

            spectrum['Contributions'] = \
                store_contributions(binning, model, output_size=output_size-3)
            if solution is not None:
                out.store_dictionary(solution, group_name='Solutions')
                priors = {}
                priors['Profiles'] = profiles
                priors['Spectra'] = spectrum
                out.store_dictionary(priors, group_name='Priors')
            else:
                out.store_dictionary(profiles, group_name='Profiles')
                out.store_dictionary(spectrum, group_name='Spectra')

            if optimizer:
                optimizer.write(o)

    root_logger.info('TauREx PROGRAM END AT %s s', datetime.datetime.now())

    if args.plot:
        wlgrid = 10000/wngrid
        if get_rank() == 0:
            import matplotlib.pyplot as plt
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)

            ax.set_xlabel(r'Wavelength $(\mu m)$')
            ax.set_ylabel(r'$(R_p/R_s)^2$')
            is_lightcurve = False
            try:
                from taurex.model.lightcurve.lightcurve import LightCurveModel
                is_lightcurve = isinstance(model, LightCurveModel)
                ax.set_xscale('linear')
            except ImportError:
                pass

            if observation is not None:
                if is_lightcurve:
                    ax.plot(observation.spectrum.flatten(),
                            label='observation')
                else:
                    ax.errorbar(observation.wavelengthGrid,
                                observation.spectrum, observation.errorBar,
                                fmt='.',
                                label='observation')

            if is_lightcurve:
                ax.plot(result[1], label='forward model')
            else:

                if inst_result is not None:
                    from taurex.util.util import wnwidth_to_wlwidth
                    inst_wngrid, inst_spectrum, \
                        inst_noise, inst_width = inst_result

                    inst_wlgrid = 10000/inst_wngrid

                    inst_wlwidth = wnwidth_to_wlwidth(inst_wngrid, inst_width)

                    ax.errorbar(inst_wlgrid, inst_spectrum, inst_noise,
                                inst_wlwidth/2, '.', label='Instrument')

                else:
                    ax.plot(wlgrid, binning.bin_model(
                        result)[1], label='forward model')

                ax.set_xscale('log')

            if args.contrib:
                native_grid, contrib_result = model.model_contrib(
                    wngrid=wngrid)

                for contrib_name, contrib in contrib_result.items():

                    flux, tau, extras = contrib

                    binned = binning.bindown(native_grid, flux)
                    if is_lightcurve:
                        ax.plot(binned[1], label=contrib_name)
                    else:
                        ax.plot(wlgrid, binned[1], label=contrib_name)

            if args.full_contrib:
                native_grid, contrib_result = model.model_full_contrib(
                    wngrid=wngrid)

                for contrib_name, contrib in contrib_result.items():

                    for name, flux, tau, extras in contrib:

                        label = '{} - {}'.format(contrib_name, name)

                        binned = binning.bindown(native_grid, flux)
                        if is_lightcurve:
                            ax.plot(binned[1], label=label)
                        else:
                            ax.plot(wlgrid, binned[1], label=label)

            plt.legend()
            plt.show()
        else:
            logging.getLogger('taurex').warning(
                'Number of processes > 1 so not plotting')


if __name__ == "__main__":

    main()



from taurex.log import Logger


class Instrument(Logger):
    """
    *Abstract class*

    Defines some method that transforms
    a spectrum and generates noise.

    """
    def __init__(self):
        super().__init__(self.__class__.__name__)

    def model_noise(self, model, model_res=None, num_observations=1):
        """
        **Requires implementation**

        For a given forward model (and optional result)
        Resample the spectrum and compute noise profile.

        Parameters
        ----------

        model: :class:`~taurex.model.model.ForwardModel`
            Forward model to pass.

        model_res: :obj:`tuple`, optional
            Result from :func:`~taurex.model.model.ForwardModel.model`

        num_observations: int, optional
            Number of observations to simulate
        """

        raise NotImplementedError



from taurex.log import Logger
import numpy as np
from astropy.constants import N_A, k_B
import astropy.units as u
from functools import lru_cache


class RadisHITRANOpacity(Logger):
    """
    This is the base class for computing opactities

    """

    def __init__(self, molecule_name, wn_start=600, wn_end=30000, wn_points=10000):
        super().__init__(self.__class__.__name__)
        import radis

        step = (wn_end-wn_start)/wn_points

        self.info('RADIS Grid set to %s %s %s', wn_start, wn_end, step)

        self.rad_xsec = radis.SpectrumFactory(wavenum_min=wn_start,
                                              wavenum_max=wn_end,
                                              isotope='1',  # 'all',
                                              # depends on HAPI benchmark.
                                              wstep=step,
                                              verbose=0,
                                              cutoff=1e-27,
                                              mole_fraction=1.0,
                                              broadening_max_width=10.0,  # Corresponds to WavenumberWingHW/HWHM=50 in HAPI
                                              molecule=molecule_name,
                                              )

        self.rad_xsec.fetch_databank('astroquery', load_energies=False)

        self._molecule_name = molecule_name

        s = self.rad_xsec.eq_spectrum(Tgas=296, pressure=1)
        wn, absnce = s.get('abscoeff')

        self.wn = 1e7/wn

    @property
    def moleculeName(self):
        return self._molecule_name

    @property
    def wavenumberGrid(self):
        return self.wn

    @property
    def temperatureGrid(self):
        raise NotImplementedError

    @property
    def pressureGrid(self):
        raise NotImplementedError

    @lru_cache(maxsize=500)
    def compute_opacity(self, temperature, pressure):

        pressure_pascal = pressure*u.Pascal
        pressure_bar = pressure_pascal.to(u.bar)
        temperature_K = temperature * u.K

        density = ((pressure_bar)/(k_B*temperature_K)).value*1000

        s = self.rad_xsec.eq_spectrum(
            Tgas=temperature, pressure=pressure_bar.value)
        wn, absnce = s.get('abscoeff')

        return absnce/density

    def opacity(self, temperature, pressure, wngrid=None):
        orig = np.nan_to_num(self.compute_opacity(temperature, pressure))

        if wngrid is None:
            return orig
        else:
            # min_max =  (self.wavenumberGrid <= wngrid.max() ) & (self.wavenumberGrid >= wngrid.min())

            # total_bins = self.wavenumberGrid[min_max].shape[0]
            # if total_bins > wngrid.shape[0]:
            #     return np.append(np.histogram(self.wavenumberGrid,wngrid, weights=orig)[0]/np.histogram(self.wavenumberGrid,wngrid)[0],0)

            # else:
            return np.interp(wngrid, self.wavenumberGrid, orig)




from .util import class_for_name
import numpy as np


def get_klass_args(klass):
    import inspect
    args, varargs, varkw, defaults = inspect.getargspec(klass.__init__)
    if defaults is None:
        return []
    keyword_args = args[-len(defaults):]

    return keyword_args


def load_generic_profile_from_hdf5(loc, module, identifier, 
                                   profile_type=None, premade_dict=None,
                                   replacement_dict=None):

    if profile_type is None:
        profile_type = loc[identifier][()]

    temp_keys = list(loc.keys())
    klass = class_for_name(module, profile_type)

    klass_kwargs = get_klass_args(klass)

    args_dict = premade_dict or {}
    repl_dict = replacement_dict or {}

    for kw in klass_kwargs:

        if kw in temp_keys:
            v = loc[kw][()]
            if isinstance(v, np.ndarray) and v.dtype.type is np.string_:
                from taurex.util.util import decode_string_array
                v = decode_string_array(v)
            if kw in repl_dict:
                args_dict[kw] = repl_dict[kw]
            else:
                args_dict[kw] = v
            

    return klass(**args_dict)


def load_temperature_from_hdf5(loc,replacement_dict=None):
    return load_generic_profile_from_hdf5(loc['Temperature'],
                                          'taurex.data.profiles.temperature',
                                          'temperature_type',
                                          replacement_dict=replacement_dict)


def load_pressure_from_hdf5(loc,replacement_dict=None):
    return load_generic_profile_from_hdf5(loc['Pressure'],
                                          'taurex.data.profiles.pressure',
                                          'pressure_type',
                                          replacement_dict=replacement_dict)


def load_gas_from_hdf5(loc, molecule, replacement_dict=None):
    return load_generic_profile_from_hdf5(loc[molecule],
                                          'taurex.data.profiles.chemistry',
                                          'gas_type',
                                          replacement_dict=replacement_dict)


def load_planet_from_hdf5(loc,replacement_dict=None):
    return load_generic_profile_from_hdf5(loc['Planet'],
                                          'taurex.data.planet','planet_type',
                                          profile_type='Planet',
                                          replacement_dict=replacement_dict)


def load_star_from_hdf5(loc,replacement_dict=None):
    return load_generic_profile_from_hdf5(loc['Star'],
                                          'taurex.data.stellar', 'star_type',
                                          replacement_dict=replacement_dict)


def load_chemistry_from_hdf5(loc,replacement_dict=None):
    from taurex.data.profiles.chemistry import TaurexChemistry
    from taurex.util.util import decode_string_array
    chemistry = load_generic_profile_from_hdf5(loc['Chemistry'],
                                               'taurex.data.profiles.chemistry',
                                               'chemistry_type',
                                          replacement_dict=replacement_dict)

    if isinstance(chemistry, TaurexChemistry):
        for mol in decode_string_array(loc['Chemistry']['active_gases'][()]):
            if mol not in chemistry._fill_gases:
                chemistry.addGas(load_gas_from_hdf5(loc['Chemistry'], mol))
        for mol in decode_string_array(loc['Chemistry']['inactive_gases']):
            if mol not in chemistry._fill_gases:
                chemistry.addGas(load_gas_from_hdf5(loc['Chemistry'], mol))

    return chemistry


def load_contrib_from_hdf5(loc, contribution,replacement_dict=None):
    return load_generic_profile_from_hdf5(loc[contribution],
                                          'taurex.contributions',
                                          'contrib_type',
                                          profile_type=contribution,
                                          replacement_dict=replacement_dict)




def load_model_from_hdf5(loc, replacement_dict=None):
    chemistry = load_chemistry_from_hdf5(loc, replacement_dict=replacement_dict)
    pressure = load_pressure_from_hdf5(loc, replacement_dict=replacement_dict)
    temperature = load_temperature_from_hdf5(loc, replacement_dict=replacement_dict)
    planet = load_planet_from_hdf5(loc, replacement_dict=replacement_dict)
    star = load_star_from_hdf5(loc, replacement_dict=replacement_dict)

    kwargs ={}

    kwargs['planet'] = planet
    kwargs['star'] = star
    kwargs['chemistry'] = chemistry
    kwargs['temperature_profile'] = temperature
    kwargs['pressure_profile'] = pressure

    model = load_generic_profile_from_hdf5(loc, 'taurex.model',  'model_type',
                                           premade_dict=kwargs,
                                           replacement_dict=replacement_dict)

    contribution_loc = loc['Contributions']

    def contrib_iterator(loc):
        import h5py
        for key in loc.keys():
            item = loc[key]
            if isinstance(item, h5py.Group):
                yield key

    for contrib in contrib_iterator(contribution_loc):
        model.add_contribution(load_contrib_from_hdf5(contribution_loc,
                                                      contrib,
                                                      replacement_dict=replacement_dict))

    return model


def taurex_hdf5_to_model(filename, replacement_dict=None):

    import h5py
    with h5py.File(filename, 'r') as f:
        model = load_model_from_hdf5(f['ModelParameters'],
                                     replacement_dict=replacement_dict)

    return model


def taurex_hdf5_to_observation(filename):

    import h5py
    from taurex.util.util import wnwidth_to_wlwidth
    from taurex.data.spectrum import ArraySpectrum
    with h5py.File(filename, 'r') as f:
        try:
            instrument_section = f['Output']['Spectra']
        except KeyError:
            raise KeyError('No instrument data found in HDF5 or retrieval hdf5 used')
        inst_wngrid = instrument_section['instrument_wngrid'][...]
        inst_spectrum = instrument_section['instrument_spectrum'][...]
        inst_noise = instrument_section['instrument_noise'][...]
        inst_width = instrument_section['instrument_wnwidth'][...]

        inst_wlgrid = 10000/inst_wngrid

        inst_wlwidth = wnwidth_to_wlwidth(inst_wngrid, inst_width)
        observation = ArraySpectrum(np.vstack([inst_wlgrid, inst_spectrum,
                                               inst_noise, inst_wlwidth]).T)
    return observation



from .isothermal import Isothermal
from .guillot import Guillot2010
from .tprofile import TemperatureProfile
from .npoint import NPoint
from .rodgers import Rodgers2000
from .file import TemperatureFile


from taurex.binning import Binner
from taurex import OutputSize


class NativeBinner(Binner):
    """
    A `do-nothing` binner. This is useful when the pipeline expects
    a binner but none is given. Simplifies implementation and also
    handles dictionary writing of the forward model.
    """

    def bindown(self, wngrid, spectrum, grid_width=None, error=None):
        """Does nothing, only returns function arguments"""
        return wngrid, spectrum, error, grid_width

    def generate_spectrum_output(self, model_output,
                                 output_size=OutputSize.heavy):
        output = {}

        wngrid, flux, tau, extra = model_output

        output['native_wngrid'] = wngrid
        output['native_wlgrid'] = 10000/wngrid
        output['native_spectrum'] = flux

        if output_size > OutputSize.light:
            output['native_tau'] = tau

        return output



from .array import ArraySpectrum
import numpy as np


class ObservedSpectrum(ArraySpectrum):
    """
    Loads an observed spectrum from a text file and computes bin
    edges and bin widths. Spectrum must be 3-4 columns with ordering:
        1. wavelength
        2. spectral data
        3. error
        4. (optional) bin width

    If no bin width is present then they are computed.

    Parameters
    -----------
    filename: string
        Path to observed spectrum file.

    """

    def __init__(self, filename):
        self._filename = filename

        super().__init__(np.loadtxt(self._filename))



from .contribution import Contribution, contribute_tau
import numpy as np
from taurex.data.fittable import fitparam
from taurex.external.mie import bh_mie


class BHMieContribution(Contribution):
    """
    Computes a Mie scattering contribution using method given by
    Bohren & Huffman 2007

    Parameters
    ----------
    mie_path: str
        Path to molecular scattering parameters

    mie_type: str of either ``cloud`` or ``haze``
        Type of mie cloud

    bh_particle_radius: float
        Radius of scattering particles in um

    bh_clouds_mix: float
        Mix ratio in atmosphere

    bh_clouds_bottomP: float
        Bottom of cloud deck in Pa

    bh_clouds_topP: float
        Top of cloud deck in Pa


    """
    def __init__(self, mie_path=None, mie_type='cloud', bh_particle_radius=1.0,
                 bh_clouds_mix=1e-6, bh_clouds_bottomP=1e0,
                 bh_clouds_topP=1e-3):
        super().__init__('Mie')
        self._mie_path = mie_path
        self.load_mie_indices()
        self._mie_type = mie_type.lower()

        self._mie_radius = bh_particle_radius
        self._mix_cloud_mix = bh_cloud_mix
        self._cloud_top_pressure = bh_clouds_topP
        self._cloud_bottom_pressure = bh_clouds_bottomP

    def load_mie_indices(self):
        import pathlib
        if self._mie_path is None:
            raise Exception('No mie file path defined')
        # loading file
        mie_raw = np.loadtxt(self._mie_path, skiprows=1)

        # saving to memory
        species_name = pathlib.Path(self._mie_path).stem
        self.info('Preloading Mie refractive indices for %s' % species_name)
        self.mie_indices = mie_raw
        self.mie_species = species_name

    @property
    def mieSpecies(self):
        return self.mie_species

    @property
    def wavelengthGrid(self):
        return self.mie_indices[:, 0]

    @property
    def wavenumberGrid(self):
        return 10000/self.wavelengthGrid

    @property
    def realReference(self):
        return np.ascontiguousarray(self.mie_indices[:, 1])

    @property
    def imaginaryReference(self):
        return np.ascontiguousarray(self.mie_indices[:, 2])

    @property
    def mieType(self):
        return self._mie_type

    @fitparam(param_name='bh_particle_radius',
              param_latex='$R^{bh}_\mathrm{clouds}$',
              default_fit=False,
              default_bounds=[-10, 1])
    def particleSize(self):
        """
        Particle size in um
        """
        return self._mie_radius

    @particleSize.setter
    def particleSize(self, value):
        self._mie_radius = value

    @fitparam(param_name='bh_clouds_topP',
              param_latex='$P^{bh}_\mathrm{top}$',
              default_fit=False,
              default_bounds=[-1, 1])
    def cloudTopPressure(self):
        """
        Pressure at top of cloud deck in Pa
        """
        return self._cloud_top_pressure

    @cloudTopPressure.setter
    def cloudTopPressure(self, value):
        self._cloud_top_pressure = value

    @fitparam(param_name='bh_clouds_bottomP',
              param_latex='$P^{bh}_\mathrm{bottom}$',
              default_fit=False,
              default_bounds=[-1, 1])
    def cloudBottomPressure(self):
        """
        Pressure at bottom of cloud deck in Pa
        """
        return self._cloud_bottom_pressure

    @cloudBottomPressure.setter
    def cloudBottomPressure(self, value):
        self._cloud_bottom_pressure = value

    @fitparam(param_name='bh_clouds_mix',
              param_latex='$\chi^{bh}_\mathrm{clouds}$',
              default_fit=False,
              default_bounds=[-1, 1])
    def cloudMixing(self):
        """
        Cloud mixing ratio
        """
        return self._mix_cloud_mix

    @cloudMixing.setter
    def cloudMixing(self, value):
        self._mix_cloud_mix = value

    def contribute(self, model, start_layer, end_layer, density_offset,
                   layer, density, tau, path_length=None):
        """
        Contributes to optical depth between cloud pressures:
        ``bh_clouds_topP`` and ``bh_clouds_bottomP``

        """
        if model.pressureProfile[layer] <= self._cloud_bottom_pressure and \
                model.pressureProfile[layer] >= self._cloud_top_pressure:

            contribute_tau(start_layer, end_layer, density_offset,
                           self.sigma_mie, density, path_length, self._nlayers,
                           self._ngrid, layer, tau)

    def build(self, model):
        """
        Preperes mie scattering cross-section

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`

        """
        wavegrid = self.wavelengthGrid*1e-4  # micron to cm
        a = self._mie_radius * 1e-4  # micron to cm
        agrid = None
        na = None
        # getting particle size distribution
        if self.mieType == 'cloud':
            # particle size distribution micron grid
            agrid = np.linspace(1e-7, a*3, 30)
            # earth clouds equ. 36 Sharp & Burrows 2007
            na = (agrid/a)**6 * np.exp((-6.0*(agrid/a)))
        elif self.mieType == 'haze':
            # particle size distribution micron grid
            agrid = np.linspace(1e-7, a*15, 50)
            # haze distributino equ. 37 Sharp & Burrows 2007
            na = agrid/a * np.exp((-2.0*(agrid/a)**0.5))
        else:
            raise Exception('Unknown Mie type {}'.format(self.mieType))
        na /= np.max(na)  # normalise into weigtings
        na_clip = na[na > 1e-3]  # curtails wings
        agrid_clip = agrid[na > 1e-3]

        # Running Mie model for particle sizes in distribution
        sig_out = np.ndarray(shape=(len(wavegrid), len(agrid_clip)))
        for i, ai in enumerate(agrid_clip):

            sig_out[:, i] = bh_mie(ai, wavegrid, self.realReference,
                                   self.imaginaryReference)

        # average mie cross section weighted by particle size distribution
        self._sig_out_aver = np.average(sig_out, weights=na_clip, axis=1)

    def prepare_each(self, model, wngrid):
        """
        Interpolates the precomputed mie scattering opacity to the desired
        wavelength grid. Yields only once

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            ``BH`` and the weighted mie opacity.

        """
        self._nlayers = model.nLayers
        self._ngrid = wngrid.shape[0]

        self.sigma_mie = np.interp(wngrid, self.wavenumberGrid,
                                   self._sig_out_aver)*self._mix_cloud_mix

        yield 'BH', self.sigma_mie

    def write(self, output):
        contrib = super().write(output)
        contrib.write_scalar('bh_cloud_particle_size', self._mie_radius)
        contrib.write_scalar('bh_cloud_topP', self._cloud_top_pressure)
        contrib.write_scalar('bh_cloud_bottomP', self._cloud_bottom_pressure)
        contrib.write_scalar('bh_cloud_mix', self._mix_cloud_mix)
        contrib.write_string('mie_path', self._mie_path)
        contrib.write_string('mie_type', self._mie_type)
        return contrib



"""
Atmospheric chemistry related modules
"""

try:
    from .acechemistry import ACEChemistry
except ImportError:
    pass


from .taurexchemistry import TaurexChemistry
from .gas.constantgas import ConstantGas
from .gas.twolayergas import TwoLayerGas
from .filechemistry import ChemistryFile



from .interpolateopacity import InterpolatingOpacity
import pickle
import numpy as np
import pathlib


class PickleOpacity(InterpolatingOpacity):
    """
    This is the base class for computing opactities

    """

    def __init__(self, filename, interpolation_mode='linear'):
        super().__init__('PickleOpacity:{}'.format(pathlib.Path(filename).stem[0:10]),
                         interpolation_mode=interpolation_mode)

        self._filename = filename
        self._molecule_name = None
        self._spec_dict = None
        self._resolution = None
        self._load_pickle_file(filename)

    @property
    def moleculeName(self):
        return self._molecule_name

    @property
    def xsecGrid(self):
        return self._xsec_grid

    def _load_pickle_file(self, filename):

        # Load the pickle file
        self.info('Loading opacity from {}'.format(filename))
        try:
            with open(filename, 'rb') as f:
                self._spec_dict = pickle.load(f)
        except UnicodeDecodeError:
            with open(filename, 'rb') as f:
                self._spec_dict = pickle.load(f, encoding='latin1')

        self._wavenumber_grid = self._spec_dict['wno']

        self._temperature_grid = self._spec_dict['t']
        self._pressure_grid = self._spec_dict['p']*1e5
        self._xsec_grid = self._spec_dict['xsecarr']
        self._resolution = np.average(np.diff(self._wavenumber_grid))
        self._molecule_name = self._spec_dict['name']

        self._min_pressure = self._pressure_grid.min()
        self._max_pressure = self._pressure_grid.max()
        self._min_temperature = self._temperature_grid.min()
        self._max_temperature = self._temperature_grid.max()
        self.clean_molecule_name()

    def clean_molecule_name(self):
        splits = self.moleculeName.split('_')
        self._molecule_name = splits[0]

    @property
    def wavenumberGrid(self):
        return self._wavenumber_grid

    @property
    def temperatureGrid(self):
        return self._temperature_grid

    @property
    def pressureGrid(self):
        return self._pressure_grid

    @property
    def resolution(self):
        return self._resolution

        # return factor*(q_11*(Pmax-P)*(Tmax-T) + q_21*(P-Pmin)*(Tmax-T) + q_12*(Pmax-P)*(T-Tmin) + q_22*(P-Pmin)*(T-Tmin))



from .tprofile import TemperatureProfile
import numpy as np
from taurex.data.fittable import fitparam


class Guillot2010(TemperatureProfile):
    """

    TP profile from Guillot 2010, A&A, 520, A27 (equation 49)
    Using modified 2stream approx.
    from Line et al. 2012, ApJ, 749,93 (equation 19)


    Parameters
    -----------
        T_irr: float
            planet equilibrium temperature
            (Line fixes this but we keep as free parameter)
        kappa_ir: float
            mean infra-red opacity
        kappa_v1: float
            mean optical opacity one
        kappa_v2: float
            mean optical opacity two
        alpha: float
            ratio between kappa_v1 and kappa_v2 downwards radiation stream

    """

    def __init__(self, T_irr=1500, kappa_irr=0.01, kappa_v1=0.005,
                 kappa_v2=0.005, alpha=0.5):
        super().__init__('Guillot')

        self.T_irr = T_irr

        self.kappa_ir = kappa_irr
        self.kappa_v1 = kappa_v1
        self.kappa_v2 = kappa_v2
        self.alpha = alpha

    @fitparam(param_name='T_irr',
              param_latex='$T_\\mathrm{irr}$',
              default_fit=True,
              default_bounds=[1300, 2500])
    def equilTemperature(self):
        """Planet equilibrium temperature"""
        return self.T_irr

    @equilTemperature.setter
    def equilTemperature(self, value):
        self.T_irr = value

    @fitparam(param_name='kappa_irr',
              param_latex='$k_\\mathrm{irr}$',
              default_fit=False,
              default_bounds=[1e-10, 1],
              default_mode='log')
    def meanInfraOpacity(self):
        """mean infra-red opacity"""
        return self.kappa_ir

    @meanInfraOpacity.setter
    def meanInfraOpacity(self, value):
        self.kappa_ir = value

    @fitparam(param_name='kappa_v1',
              param_latex='$k_\\mathrm{1}$',
              default_fit=False,
              default_bounds=[1e-10, 1],
              default_mode='log')
    def meanOpticalOpacity1(self):
        """mean optical opacity one"""
        return self.kappa_v1

    @meanOpticalOpacity1.setter
    def meanOpticalOpacity1(self, value):
        self.kappa_v1 = value

    @fitparam(param_name='kappa_v2',
              param_latex='$k_\\mathrm{2}$',
              default_fit=False,
              default_bounds=[1e-10, 1],
              default_mode='log')
    def meanOpticalOpacity2(self):
        """mean optical opacity two"""
        return self.kappa_v2

    @meanOpticalOpacity2.setter
    def meanOpticalOpacity2(self, value):
        self.kappa_v2 = value

    @fitparam(param_name='alpha', param_latex='$\\alpha$',
              default_fit=False, default_bounds=[0.0, 1.0])
    def opticalRatio(self):
        """ratio between kappa_v1 and kappa_v2 """
        return self.alpha

    @opticalRatio.setter
    def opticalRatio(self, value):
        self.alpha = value

    @property
    def profile(self):
        """

        Returns a guillot temperature temperature profile

        Returns
        --------

        temperature_profile : :obj:np.array(float)

        """

        planet_grav = self.planet.gravity
        gamma_1 = self.kappa_v1/(self.kappa_ir + 1e-10)
        gamma_2 = self.kappa_v2/(self.kappa_ir + 1e-10)
        tau = self.kappa_ir * self.pressure_profile / planet_grav

        T_int = 100  # todo internal heat parameter looks suspicious..

        def eta(gamma, tau):
            import scipy.special as spe

            part1 = 2.0/3.0 + 2.0 / (3.0*gamma) * (1.0 +
                                                   (gamma*tau/2.0 - 1.0) *
                                                   np.exp(-1.0 * gamma * tau))

            part2 = 2.0 * gamma / 3.0 * (1.0 - tau**2/2.0) * \
                spe.expn(2, (gamma * tau))

            return part1 + part2

        T4 = 3.0 * T_int**4/4.0 * (2.0/3.0 + tau) + 3.0*self.T_irr**4 / 4.0 * \
            (1.0 - self.alpha) * eta(gamma_1, tau) + \
            3.0 * self.T_irr**4/4.0 * self.alpha * eta(gamma_2, tau)

        T = T4**0.25

        return T

    def write(self, output):
        temperature = super().write(output)
        temperature.write_scalar('T_irr', self.T_irr)
        temperature.write_scalar('kappa_irr', self.kappa_ir)
        temperature.write_scalar('kappa_v1', self.kappa_v1)
        temperature.write_scalar('kappa_v2', self.kappa_v2)
        temperature.write_scalar('alpha', self.alpha)
        return temperature



from taurex.log import Logger, disableLogging, enableLogging
import numpy as np
import math
from taurex import OutputSize


class Optimizer(Logger):
    """
    A base class that handles fitting and optimization of forward models.
    The class handles the compiling and management of fitting parameters in 
    forward models, in its current form it cannot fit and requires a class 
    derived from it to implement the :func:`compute_fit` function.

    Parameters
    ----------
    name:  str
        Name to be used in logging

    observed : :class:`~taurex.data.spectrum.spectrum.BaseSpectrum`, optional
        See :func:`set_observed`

    model : :class:`~taurex.model.model.ForwardModel`, optional
        See :func:`set_model`

    sigma_fraction: float, optional
        Fraction of weights to use in computing the error. (Default: 0.1)

    """

    def __init__(self, name, observed=None, model=None, sigma_fraction=0.1):
        super().__init__(name)

        self.set_model(model)
        self.set_observed(observed)
        self._model_callback = None
        self._sigma_fraction = sigma_fraction

    def set_model(self, model):
        """
        Sets the model to be optimized/fit

        Parameters
        ----------
        model : :class:`~taurex.model.model.ForwardModel`
            The forward model we wish to optimize

        """
        self._model = model

    def set_observed(self, observed):
        """
        Sets the observation to optimize the model to

        Parameters
        ----------
        observed : :class:`~taurex.data.spectrum.spectrum.BaseSpectrum`
            Observed spectrum we will optimize to

        """
        self._observed = observed
        if observed is not None:
            self._binner = observed.create_binner()

    def compile_params(self):
        """ 


        Goes through and compiles all parameters within the model that
        we will be retrieving. Called before :func:`compute_fit`


        """
        self.info('Initializing parameters')
        self.fitting_parameters = []
        # param_name,param_latex,
        #                 fget.__get__(self),fset.__get__(self),
        #                         default_fit,default_bounds
        for params in self._model.fittingParameters.values():
            name, latex, fget, fset, mode, to_fit, bounds = params

            self.debug('Checking fitting parameter {}'.format(params))
            if to_fit:
                self.fitting_parameters.append(params)

        self.info('-------FITTING---------------')
        self.info('Parameters to be fit:')
        for params in self.fitting_parameters:
            name, latex, fget, fset, mode, to_fit, bounds = params
            self.info('{}: Value: {} Mode:{} Boundaries:{}'.format(
                name, fget(), mode, bounds))

    def update_model(self, fit_params):
        """
        Updates the model with new parameters

        Parameters
        ----------
        fit_params : :obj:`list`
            A list of new values to apply to the model. The list of values are
            assumed to be in the same order as the parameters given by 
            :func:`fit_names`


        """

        for value, param in zip(fit_params, self.fitting_parameters):
            name, latex, fget, fset, mode, to_fit, bounds = param
            if mode == 'log':
                value = 10**value
            fset(value)

    @property
    def fit_values_nomode(self):
        """ 

        Returns a list of the current values of a fitting parameter. 
        Regardless of the ``mode`` setting

        Returns
        -------
        :obj:`list`:
            List of each value of a fitting parameter

        """

        return [c[2]() for c in self.fitting_parameters]

    @property
    def fit_values(self):
        """ 

        Returns a list of the current values of a fitting parameter. This 
        respects the ``mode`` setting

        Returns
        -------
        :obj:`list`:
            List of each value of a fitting parameter

        """

        return [c[2]() if c[4] == 'linear' else math.log10(c[2]()) 
                for c in self.fitting_parameters]

    @property
    def fit_boundaries(self):
        """

        Returns the fitting boundaries of the parameter

        Returns
        -------
        :obj:`list`:
            List of boundaries for each fitting parameter. It takes the form of
            a python :obj:`tuple` with the form 
            ( ``bound_min`` , ``bound_max`` )

        """
        return [c[-1] if c[4] == 'linear'
                else (math.log10(c[-1][0]), math.log10(c[-1][1]))
                for c in self.fitting_parameters]

    @property
    def fit_names(self):
        """

        Returns the names of the model parameters we will be fitting

        Returns
        -------
        :obj:`list`:
            List of names of parameters that will be fit

        """
        return [c[0] if c[4] == 'linear' else 'log_{}'.format(c[0])
                for c in self.fitting_parameters]

    @property
    def fit_latex(self):
        """

        Returns the names of the parameters in LaTeX format

        Returns
        -------
        :obj:`list`
            List of parameter names in LaTeX format


        """

        return [c[1] if c[4] == 'linear' else 'log({})'.format(c[1])
                for c in self.fitting_parameters]

    def enable_fit(self, parameter):
        """
        Enables fitting of the parameter

        Parameters
        ----------
        parameter : str
            Name of the parameter we want to fit

        """

        name, latex, fget, fset, mode, to_fit, bounds = \
            self._model.fittingParameters[parameter]

        to_fit = True

        self._model.fittingParameters[parameter] = (
            name, latex, fget, fset, mode, to_fit, bounds)

    def disable_fit(self, parameter):
        """
        Disables fitting of the parameter

        Parameters
        ----------
        parameter : str
            Name of the parameter we do not want to fit

        """
        name, latex, fget, fset, mode, to_fit, bounds = \
            self._model.fittingParameters[parameter]

        to_fit = False

        self._model.fittingParameters[parameter] = (
            name, latex, fget, fset, mode, to_fit, bounds)

    def set_boundary(self, parameter, new_boundaries):
        """
        Sets the boundary of the parameter

        Parameters
        ----------
        parameter : str
            Name of the parameter we want to change

        new_boundaries : tuple of float
            New fitting boundaries, with the form
            ( ``bound_min`` , ``bound_max`` ). These should
            not take into account the ``mode`` setting of a fitting parameter.


        """
        name, latex, fget, fset, mode, to_fit, bounds = \
            self._model.fittingParameters[parameter]

        bounds = new_boundaries

        self._model.fittingParameters[parameter] = (
            name, latex, fget, fset, mode, to_fit, bounds)

    def set_factor_boundary(self, parameter, factors):
        """
        Sets the boundary of the parameter based on a factor

        Parameters
        ----------
        parameter : str
            Name of the parameter we want to change

        factor : tuple of float
            To be written



        """

        name, latex, fget, fset, mode, to_fit, bounds = \
            self._model.fittingParameters[parameter]

        value = fget()

        new_boundaries = factors[0]*value, factors[1]*value

        bounds = new_boundaries
        self._model.fittingParameters[parameter] = (
            name, latex, fget, fset, mode, to_fit, bounds)

    def set_mode(self, parameter, new_mode):
        """
        Sets the fitting mode of a parameter

        Parameters
        ----------
        parameter : str
            Name of the parameter we want to change

        new_mode : ``linear`` or ``log``
            Sets whether the parameter is fit in linear or log space



        """
        new_mode = new_mode.lower()

        name, latex, fget, fset, mode, to_fit, bounds = \
            self._model.fittingParameters[parameter]

        if new_mode not in ('log', 'linear',):
            self.error('Incorrect mode set for fit parameter,')
            raise ValueError

        self._model.fittingParameters[parameter] = (
            name, latex, fget, fset, new_mode, to_fit, bounds)

    def chisq_trans(self, fit_params, data, datastd):
        """

        Computes the Chi-Squared between the forward model and
        observation. The steps taken are:
            1. Forward model (FM) is updated with :func:`update_model`
            2. FM is then computed at its native grid then binned.
            3. Chi-squared between FM and observation is computed


        Parameters
        ----------
        fit_params : list of parameter values
            List of new parameter values to update the model

        data : obj:`ndarray`
            Observed spectrum

        datastd : obj:`ndarray`
            Observed spectrum error


        Returns
        -------
        float :
            chi-squared


        """
        from taurex.exceptions import InvalidModelException
        self.update_model(fit_params)

        obs_bins = self._observed.wavenumberGrid

        try:
            _, final_model, _, _ = self._binner.bin_model(
                self._model.model(wngrid=obs_bins))
        except InvalidModelException:
            return 1e100

        res = (data.ravel() - final_model.ravel()) / datastd.ravel()
        res = np.nansum(res*res)
        if res == 0:
            res = np.nan

        return res

    def compute_fit(self):
        """
        Unimplemented. When inheriting this should be overwritten
        to perform the actual fit.

        Raises
        ------
        NotImplementedError
            Raised when a derived class does override this function

        """
        raise NotImplementedError

    def fit(self, output_size=OutputSize.heavy):
        import time
        """

        Performs fit.

        """
        from tabulate import tabulate
        self.compile_params()

        fit_names = self.fit_names
        fit_boundaries = self.fit_boundaries

        fit_min = [x[0] for x in fit_boundaries]
        fit_max = [x[1] for x in fit_boundaries]

        fit_values = self.fit_values
        self.info('')
        self.info('-------------------------------------')
        self.info('------Retrieval Parameters-----------')
        self.info('-------------------------------------')
        self.info('')
        self.info('Dimensionality of fit: %s', len(fit_names))
        self.info('')

        output = tabulate(zip(fit_names, fit_values, fit_min, fit_max),
                          headers=['Param', 'Value', 'Bound-min', 'Bound-max'])

        self.info('\n%s\n\n', output)
        self.info('')
        start_time = time.time()
        disableLogging()
        # Compute fit here
        self.compute_fit()

        enableLogging()
        end_time = time.time()
        self.info('Sampling time %s s', end_time-start_time)
        solution = self.generate_solution(output_size=output_size)
        self.info('')
        self.info('-------------------------------------')
        self.info('------Final results------------------')
        self.info('-------------------------------------')
        self.info('')
        self.info('Dimensionality of fit: %s', len(fit_names))
        self.info('')
        for idx, optimized_map, optimized_median, values in self.get_solution():
            self.info('\n%s', '---Solution {}------'.format(idx))
            output = tabulate(zip(fit_names, optimized_map, optimized_median), headers=[
                              'Param', 'MAP', 'Median'])
            self.info('\n%s\n\n', output)
        return solution

    def write_optimizer(self, output):
        """

        Writes optimizer settings under the 'Optimizer' heading in an
        output file

        Parameters
        ----------
        output:  :class:`~taurex.output.output.Output` or :class:`~taurex.output.output.OutputGroup`
            Group (or root) in output file to write to

        Returns
        -------
        :class:`~taurex.output.output.Output` or :class:`~taurex.output.output.OutputGroup`
            Group (or root) in output file written to

        """
        output.write_string('optimizer', self.__class__.__name__)
        output.write_string_array('fit_parameter_names', self.fit_names)
        output.write_string_array('fit_parameter_latex', self.fit_latex)
        output.write_array('fit_boundary_low', np.array(
            [x[0] for x in self.fit_boundaries]))
        output.write_array('fit_boundary_high', np.array(
            [x[1] for x in self.fit_boundaries]))
        return output

    def write_fit(self, output):
        """
        Writes basic fitting parameters into output

        Parameters
        ----------
        output : :class:`~taurex.output.output.Output` or :class:`~taurex.output.output.OutputGroup`
            Group (or root) in output file to write to

        Returns
        -------
        :class:`~taurex.output.output.Output` or :class:`~taurex.output.output.OutputGroup`
            Group (or root) in output file written to

        """
        fit = output.create_group('FitParams')
        fit.write_string('fit_format', self.__class__.__name__)
        fit.write_string_array('fit_parameter_names', self.fit_names)
        fit.write_string_array('fit_parameter_latex', self.fit_latex)
        fit.write_array('fit_boundary_low', np.array(
            [x[0] for x in self.fit_boundaries]))
        fit.write_array('fit_boundary_high', np.array(
            [x[1] for x in self.fit_boundaries]))

        # This is the last sampled value ... should not be recorded to avoid confusion.
        # fit.write_list('fit_parameter_values',self.fit_values)
        # fit.write_list('fit_parameter_values_nomode',self.fit_values_nomode)
        return output

    def generate_profiles(self, solution, binning):
        from taurex import mpi
        """Generates sigma plots for profiles"""

        sample_list = None

        if mpi.get_rank() == 0:
            sample_list = list(self.sample_parameters(solution))

        sample_list = mpi.broadcast(sample_list)

        self.debug('We all got %s', sample_list)

        self.info('------------Variance generation step------------------')

        self.info('We are sampling %s points for the profiles',
                  len(sample_list))

        rank = mpi.get_rank()
        size = mpi.nprocs()

        enableLogging()

        self.info('I will only iterate through partitioned %s '
                  'points (the rest is in parallel)', len(sample_list)//size)

        disableLogging()
        count = 0

        def sample_iter():
            for parameters, weight in sample_list[rank::size]:
                self.update_model(parameters)
                enableLogging()
                if rank == 0 and count % 10 == 0 and count > 0:

                    self.info('Progress {}%'.format(
                        count*100.0 / (len(sample_list)/size)))
                disableLogging()
                yield weight

        return self._model.compute_error(sample_iter, wngrid=binning,
                                         binner=self._binner)

    def generate_solution(self, output_size=OutputSize.heavy):
        """
        Generates a dictionary with all solutions and other useful parameters
        """
        from taurex.util.output import generate_profile_dict, \
            store_contributions

        solution_dict = {}

        self.info('Generating spectra and profiles')

        # Loop through each solution, grab optimized parameters and anything 
        # else we want to store
        for solution, optimized_map, \
                optimized_median, values in self.get_solution():

            self.info('Computing solution %s', solution)
            sol_values = {}

            # Include extra stuff we might want to store (provided by the child)
            for k, v in values:
                sol_values[k] = v

            # Update the model with optimized map values
            self.update_model(optimized_map)

            opt_result = self._model.model(cutoff_grid=False)  # Run the model

            sol_values['Spectra'] = self._binner.generate_spectrum_output(
                opt_result, output_size=output_size)

            sol_values['Spectra']['Contributions'] = store_contributions(
                self._binner, self._model, output_size=output_size-3)

            # Update with the optimized median
            self.update_model(optimized_median)

            self._model.model(cutoff_grid=False)

            # Store profiles here
            sol_values['Profiles'] = generate_profile_dict(self._model)
            profile_dict, spectrum_dict = self.generate_profiles(
                solution, self._observed.wavenumberGrid)

            for k, v in profile_dict.items():
                sol_values['Profiles'][k] = v

            for k, v in spectrum_dict.items():
                sol_values['Spectra'][k] = v

            solution_dict['solution{}'.format(solution)] = sol_values

        # Compute mu derived
        for solution, optimized_map, \
                optimized_median, values in self.get_solution():

            mu = self.compute_mu_derived_trace(solution)

            solution_dict['solution{}'.format(
                solution)]['fit_params']['mu_derived'] = mu

        return solution_dict

    def compute_mu_derived_trace(self, solution):
        from taurex.util.util import quantile_corner
        from taurex.constants import AMU
        sigma_frac = self._sigma_fraction
        self._sigma_fraction = 1.0
        mu_trace = []
        weights = []
        self.info('Computing derived mu......')
        disableLogging()
        for parameters, weight in self.sample_parameters(solution):
            self.update_model(parameters)
            self._model.initialize_profiles()
            mu_trace.append(self._model.chemistry.muProfile[0]/AMU)
            weights.append(weight)
        enableLogging()

        self.info('Done!')

        self._sigma_fraction = sigma_frac

        q_16, q_50, q_84 = \
            quantile_corner(np.array(mu_trace), [0.16, 0.5, 0.84],
                            weights=np.array(weights))

        mean = np.average(mu_trace, weights=weights, axis=0)

        mu_derived = {
            'value': q_50,
            'sigma_m': q_50-q_16,
            'sigma_p': q_84-q_50,
            'trace': mu_trace,
            'mean': mean
        }
        return mu_derived

    def sample_parameters(self, solution):
        """
        **Requires implementation***

        Read traces and weights and return
        a random ``sigma_fraction`` sample of them

        Parameters
        ----------
        solution:
            a solution output from sampler

        Yields
        ------
        traces: :obj:`array`
            Traces of a particular sample

        weight: float
            Weight of sample

        """
        raise NotImplementedError

    def get_solution(self):
        """
        ** Requires implementation **

        Generator for solutions and their
        median and MAP values

        Yields
        ------

        solution_no: int
            Solution number
        
        map: :obj:`array`
            Map values
        
        median: :obj:`array`
            Median values
        
        extra: :obj:`list`
            List of tuples of extra information to store.
            Must be of form ``(name, data)``
            
        

        """
        raise NotImplementedError

    def write(self, output):
        """
        Creates 'Optimizer'
        them respectively



        Parameters
        ----------
        output : :class:`~taurex.output.output.Output` or :class:`~taurex.output.output.OutputGroup`
            Group (or root) in output file to write to



        """
        opt = output.create_group('Optimizer')
        self.write_optimizer(opt)



from .gas import Gas
from taurex.util import movingaverage, molecule_texlabel
import numpy as np


class TwoLayerGas(Gas):
    """

    Two layer gas profile.

    A gas profile with two different mixing layers at the surface of the
    planet and top of the atmosphere seperated at a defined
    pressure point and smoothened.

    Parameters
    -----------
    molecule_name : str
        Name of molecule

    mix_ratio_surface : float
        Mixing ratio of the molecule on the planet surface

    mix_ratio_top : float
        Mixing ratio of the molecule at the top of the atmosphere

    mix_ratio_P : float
        Boundary Pressure point between the two layers

    mix_ratio_smoothing : float , optional
        smoothing window

    """

    def __init__(self, molecule_name='CH4', mix_ratio_surface=1e-4,
                 mix_ratio_top=1e-8, mix_ratio_P=1e3,
                 mix_ratio_smoothing=10):
        super().__init__(self.__class__.__name__, molecule_name=molecule_name)

        self._mix_surface = mix_ratio_surface
        self._mix_top = mix_ratio_top
        self._mix_ratio_pressure = mix_ratio_P
        self._mix_ratio_smoothing = mix_ratio_smoothing
        self._mix_profile = None
        self.add_surface_param()
        self.add_top_param()
        self.add_P_param()

    @property
    def mixProfile(self):
        """

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer

        """
        return self._mix_profile

    @property
    def mixRatioSurface(self):
        """Abundance on the planets surface"""
        return self._mix_surface

    @property
    def mixRatioTop(self):
        """Abundance on the top of atmosphere"""
        return self._mix_top

    @property
    def mixRatioPressure(self):
        return self._mix_ratio_pressure

    @property
    def mixRatioSmoothing(self):
        return self._mix_ratio_smoothing

    @mixRatioSurface.setter
    def mixRatioSurface(self, value):
        self._mix_surface = value

    @mixRatioTop.setter
    def mixRatioTop(self, value):
        self._mix_top = value

    @mixRatioPressure.setter
    def mixRatioPressure(self, value):
        self._mix_pressure = value

    @mixRatioSmoothing.setter
    def mixRatioSmoothing(self, value):
        self._mix_smoothing = value

    def add_surface_param(self):
        """
        Generates surface fitting parameters. Has the form
        ''Moleculename_surface'
        """
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_surface = '{}_surface'.format(param_name)
        param_surf_tex = '{}_surface'.format(param_tex)

        def read_surf(self):
            return self._mix_surface

        def write_surf(self, value):
            self._mix_surface = value

        fget_surf = read_surf
        fset_surf = write_surf

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(param_surface, param_surf_tex, fget_surf,
                                fset_surf, 'log', default_fit, bounds)

    def add_top_param(self):
        """
        Generates TOA fitting parameters. Has the form:
        'Moleculename_top'
        """
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_top = '{}_top'.format(param_name)
        param_top_tex = '{}_top'.format(param_tex)

        def read_top(self):
            return self._mix_top

        def write_top(self, value):
            self._mix_top = value

        fget_top = read_top
        fset_top = write_top

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(param_top, param_top_tex, fget_top,
                                fset_top, 'log', default_fit, bounds)

    def add_P_param(self):
        """
        Generates pressure fitting parameter. Has the form
        'Moleculename_P'
        """
        mol_name = self.molecule
        mol_tex = molecule_texlabel(mol_name)

        param_P = '{}_P'.format(mol_name)
        param_P_tex = '{}_P'.format(mol_tex)

        def read_P(self):
            return self._mix_ratio_pressure

        def write_P(self, value):
            self._mix_ratio_pressure = value

        fget_P = read_P
        fset_P = write_P

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(param_P, param_P_tex,
                                fget_P, fset_P, 'log', default_fit, bounds)

    def initialize_profile(self, nlayers=None, temperature_profile=None,
                           pressure_profile=None, altitude_profile=None):
        self._mix_profile = np.zeros(nlayers)

        smooth_window = self._mix_ratio_smoothing
        P_layer = np.abs(pressure_profile - self._mix_ratio_pressure).argmin()

        start_layer = max(int(P_layer-smooth_window/2), 0)

        end_layer = min(int(P_layer+smooth_window/2), nlayers-1)

        Pnodes = [pressure_profile[0], pressure_profile[start_layer],
                  pressure_profile[end_layer], pressure_profile[-1]]

        Cnodes = [self.mixRatioSurface, self.mixRatioSurface,
                  self.mixRatioTop, self.mixRatioTop]

        chemprofile = 10**np.interp((np.log(pressure_profile[::-1])),
                                    np.log(Pnodes[::-1]),
                                    np.log10(Cnodes[::-1]))

        wsize = nlayers * (smooth_window / 100.0)

        if (wsize % 2 == 0):
            wsize += 1

        C_smooth = 10**movingaverage(np.log10(chemprofile), int(wsize))

        border = np.int((len(chemprofile) - len(C_smooth)) / 2)

        self._mix_profile = chemprofile[::-1]

        self._mix_profile[border:-border] = C_smooth[::-1]

    def write(self, output):
        gas_entry = super().write(output)
        gas_entry.write_scalar('mix_ratio_top', self.mixRatioTop)
        gas_entry.write_scalar('mix_ratio_surface', self.mixRatioSurface)
        gas_entry.write_scalar('mix_ratio_P', self.mixRatioPressure)
        gas_entry.write_scalar('mix_ratio_smoothing', self.mixRatioSmoothing)

        return gas_entry



from taurex.log import Logger
from taurex.constants import G, RJUP, MJUP, AU
from .fittable import fitparam, Fittable
from taurex.output.writeable import Writeable


class Planet(Fittable, Logger, Writeable):
    """Holds information on a planet and its properties and
    derived properties

    Parameters
    -----------

    planet_mass: float, optional
        mass in terms of Jupiter mass of the planet
    planet_radius: float, optional
        radius in terms of Jupiter radii of the planet
    planet_distance: float, optional
        Semi-major axis in AU
    impact_param: float, optional
        Impact parameter
    orbital_period: float, optional
        Orbital period in days
    albedo: float, optional
        Planetary albedo
    transit_time: float, optional
        Transit time in seconds

    """

    def __init__(self, planet_mass=1.0, planet_radius=1.0,
                 planet_distance=1,
                 impact_param=0.5, orbital_period=2.0, albedo=0.3,
                 transit_time=3000.0):
        Logger.__init__(self, 'Planet')
        Fittable.__init__(self)
        self._mass = planet_mass*MJUP
        self._radius = planet_radius*RJUP
        self._distance = planet_distance*AU
        self._impact = impact_param
        self._orbit_period = orbital_period
        self._albedo = albedo
        self._transit_time = transit_time

    @fitparam(param_name='planet_mass', param_latex='$M_p$',
              default_fit=False, default_bounds=[0.5, 1.5])
    def mass(self):
        """
        Planet mass in Jupiter mass
        """
        return self._mass/MJUP

    @mass.setter
    def mass(self, value):
        self._mass = value*MJUP

    @fitparam(param_name='planet_radius', param_latex='$R_p$',
              default_fit=True, default_bounds=[0.9, 1.1])
    def radius(self):
        """
        Planet radius in Jupiter radii
        """
        return self._radius/RJUP

    @radius.setter
    def radius(self, value):
        self._radius = value*RJUP

    @property
    def fullRadius(self):
        """
        Planet radius in metres
        """
        return self._radius

    @property
    def fullMass(self):
        """
        Planet mass in kg
        """
        return self._mass

    @property
    def impactParameter(self):
        return self._impact

    @property
    def orbitalPeriod(self):
        return self._orbit_period

    @property
    def albedo(self):
        return self._albedo

    @property
    def transitTime(self):
        return self._transit_time

    @fitparam(param_name='planet_distance', param_latex='$D_{planet}$',
              default_fit=False, default_bounds=[1, 2])
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, value):
        self._distance = value

    @property
    def gravity(self):
        """
        Surface gravity in ms-2
        """
        return (G * self.fullMass) / (self.fullRadius**2)

    def gravity_at_height(self, height):
        """
        Gravity at height (m) from planet in ms-2

        Parameters
        ----------
        height: float
            Height in metres from planet surface

        Returns
        -------
        g: float
            Gravity in ms-2

        """

        return (G * self.fullMass) / ((self.fullRadius+height)**2)

    def write(self, output):
        planet = output.create_group('Planet')

        planet.write_string('planet_type', self.__class__.__name__)
        planet.write_scalar('planet_mass', self._mass/MJUP)
        planet.write_scalar('planet_radius', self._radius/RJUP)
        planet.write_scalar('planet_distance', self._distance/AU)
        planet.write_scalar('impact_param', self._impact)
        planet.write_scalar('orbital_period', self.orbitalPeriod)
        planet.write_scalar('albedo', self.albedo)
        planet.write_scalar('transit_time', self.transitTime)

        planet.write_scalar('mass_kg', self.mass)
        planet.write_scalar('radius_m', self.radius)
        planet.write_scalar('surface_gravity', self.gravity)
        return planet


class Earth(Planet):
    """An implementation for earth"""

    def __init__(self):
        Planet.__init__(self, 5.972e24, 6371000)


class Mars(Planet):

    def __init__(self):
        import astropy.units as u

        radius = (0.532*u.R_earth).to(u.jupiterRad)
        mass = (0.107*u.M_earth).to(u.jupiterMass)
        distance = 1.524
        Planet.__init__(mass=mass.value, radius=radius.value,
                        distance=distance)



from .gas import Gas
from taurex.util import molecule_texlabel
import numpy as np


class ConstantGas(Gas):
    """

    Constant gas profile. Molecular abundace is constant at each layer of the
    atmosphere

    Parameters
    -----------
    molecule_name : str
        Name of molecule

    mix_ratio : float
        Mixing ratio of the molecule

    """

    def __init__(self, molecule_name='H2O', mix_ratio=1e-5):
        super().__init__('ConstantGas', molecule_name)
        self._mix_ratio = mix_ratio
        self.add_active_gas_param()

    @property
    def mixProfile(self):
        """

        Mixing profile

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer

        """

        return self._mix_array

    def initialize_profile(self, nlayers=None, temperature_profile=None,
                           pressure_profile=None, altitude_profile=None):
        self._mix_array = self._mix_ratio*np.ones(nlayers)

    def add_active_gas_param(self):
        """
        Adds the mixing ratio as a fitting parameter
        as the name of the molecule
        """

        mol_name = self.molecule
        param_name = self.molecule
        param_tex = molecule_texlabel(mol_name)

        def read_mol(self):
            return self._mix_ratio

        def write_mol(self, value):
            self._mix_ratio = value

        fget = read_mol
        fset = write_mol

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(param_name, param_tex, fget, fset,
                                'log', default_fit, bounds)

    def write(self, output):
        gas_entry = super().write(output)
        gas_entry.write_scalar('mix_ratio', self._mix_ratio)

        return gas_entry




# -*- mode: python -*-
#!python
#cython: language_level=3
#cython: boundscheck=False
#cython: wraparound=False
# Comments above are special. Please do not remove.
cimport numpy as np  # needed for function arguments
import numpy as np # needed for np.empty_like


ctypedef np.float32_t float_t
ctypedef np.float64_t double_t
ctypedef np.int32_t int_t

cdef extern from "bhmie_lib.h":
    void compute_sigma_mie(
            const double a, 
            const int nwgrid,   
            const double * wavegrid,
            const double * ref_real, 
            const double * ref_imag,
            void * sigma_out        
            );

def bh_mie(double particle_radius, 
          np.ndarray[dtype=double_t, ndim=1, mode="c"] wavegrid,
          np.ndarray[dtype=double_t, ndim=1, mode="c"] ref_real,
          np.ndarray[dtype=double_t, ndim=1, mode="c"] ref_imag,
          ):

    cdef np.ndarray[dtype=double_t, ndim=1, mode="c"] output
    output = np.empty_like(wavegrid, dtype='d')

    if not (wavegrid.shape[0] == ref_real.shape[0]):
        raise ValueError("wave grid and ref_real shapes are not consistent")

    if not (ref_real.shape[0] == ref_imag.shape[0]):
        raise ValueError("ref_real and ref_imag shapes are not consistent")

    compute_sigma_mie(particle_radius, wavegrid.shape[0],<double*>wavegrid.data, <double*>ref_real.data,<double*>ref_imag.data, <double*>output.data)

    return output


"""
Modules that deal with computing contributions to optical depth
"""
from .contribution import Contribution, contribute_tau
from .absorption import AbsorptionContribution
from .cia import CIAContribution, contribute_cia
from .rayleigh import RayleighContribution
from .simpleclouds import SimpleCloudsContribution
from .leemie import LeeMieContribution
from .flatmie import FlatMieContribution
try:
    from .bhmie import BHMieContribution
except ImportError:
    from taurex.log.logger import root_logger
    root_logger.error('MIE could not be loaded')



from taurex.log import Logger
import numpy as np
from .opacity import Opacity
from taurex.util.math import *


class InterpolatingOpacity(Opacity):
    """
    Provides interpolation methods

    """

    def __init__(self, name, interpolation_mode='linear'):
        super().__init__(name)
        self._interp_mode = interpolation_mode

    @property
    def xsecGrid(self):
        raise NotImplementedError

    def find_closest_index(self, T, P):
        t_min = self.temperatureGrid.searchsorted(T, side='right')-1
        t_max = t_min+1

        p_min = self.pressureGrid.searchsorted(P, side='right')-1
        p_max = p_min+1

        return t_min, t_max, p_min, p_max

    def set_interpolation_mode(self, interp_mode):
        self._interp_mode = interp_mode.strip()

    def interp_temp_only(self, T, t_idx_min, t_idx_max, P, filt):
        Tmax = self.temperatureGrid[t_idx_max]
        Tmin = self.temperatureGrid[t_idx_min]
        fx0 = self.xsecGrid[P, t_idx_min, filt]
        fx1 = self.xsecGrid[P, t_idx_max, filt]

        if self._interp_mode == 'linear':
            return interp_lin_only(fx0, fx1, T, Tmin, Tmax)
        elif self._interp_mode == 'exp':
            return interp_exp_only(fx0, fx1, T, Tmin, Tmax)
        else:
            raise ValueError(
                'Unknown interpolation mode {}'.format(self._interp_mode))

    def interp_pressure_only(self, P, p_idx_min, p_idx_max, T, filt):
        Pmax = self.pressureGrid[p_idx_max]
        Pmin = self.pressureGrid[p_idx_min]
        fx0 = self.xsecGrid[p_idx_min, T, filt]
        fx1 = self.xsecGrid[p_idx_max, T, filt]

        return interp_lin_only(fx0, fx1, P, Pmin, Pmax)

    def interp_bilinear_grid(self, T, P, t_idx_min, t_idx_max, p_idx_min, p_idx_max, wngrid_filter=None):

        self.debug('Interpolating %s %s %s %s %s %s', T, P,
                   t_idx_min, t_idx_max, p_idx_min, p_idx_max)

        if p_idx_max == 0 and t_idx_max == 0:

            return np.zeros_like(self.xsecGrid[0, 0, wngrid_filter]).ravel()

        check_pressure_max = P >= self._max_pressure
        check_temperature_max = T >= self._max_temperature

        check_pressure_min = P < self._min_pressure
        check_temperature_min = T < self._min_temperature

        self.debug('Check pressure min/max %s/%s',
                   check_pressure_min, check_pressure_max)
        self.debug('Check temeprature min/max %s/%s',
                   check_temperature_min, check_temperature_max)
        # Are we both max?
        if check_pressure_max and check_temperature_max:
            self.debug('Maximum Temperature pressure reached. Using last')
            return self.xsecGrid[-1, -1, wngrid_filter].ravel()

        # Max pressure
        if check_pressure_max:
            self.debug('Max pressure reached. Interpolating temperature only')
            return self.interp_temp_only(T, t_idx_min, t_idx_max, -1, wngrid_filter)

        # Max temperature
        if check_temperature_max:
            self.debug('Max temperature reached. Interpolating pressure only')
            return self.interp_pressure_only(P, p_idx_min, p_idx_max, -1, wngrid_filter)

        if check_pressure_min and check_temperature_min:
            return self.xsecGrid[0, 0, wngrid_filter].ravel()

        if check_pressure_min:
            self.debug('Min pressure reached. Interpolating temperature only')
            return self.interp_temp_only(T, t_idx_min, t_idx_max, 0, wngrid_filter).ravel()

        if check_temperature_min:
            self.debug('Min temeprature reached. Interpolating pressure only')
            return self.interp_pressure_only(P, p_idx_min, p_idx_max, 0, wngrid_filter).ravel()

        q_11 = self.xsecGrid[p_idx_min, t_idx_min][wngrid_filter].ravel()
        q_12 = self.xsecGrid[p_idx_min, t_idx_max][wngrid_filter].ravel()
        q_21 = self.xsecGrid[p_idx_max, t_idx_min][wngrid_filter].ravel()
        q_22 = self.xsecGrid[p_idx_max, t_idx_max][wngrid_filter].ravel()

        Tmax = self.temperatureGrid[t_idx_max]
        Tmin = self.temperatureGrid[t_idx_min]
        Pmax = self.pressureGrid[p_idx_max]
        Pmin = self.pressureGrid[p_idx_min]

        if self._interp_mode == 'linear':
            return intepr_bilin(q_11, q_12, q_21, q_22, T, Tmin, Tmax, P, Pmin, Pmax)
        elif self._interp_mode == 'exp':
            return interp_exp_and_lin(q_11, q_12, q_21, q_22, T, Tmin, Tmax, P, Pmin, Pmax)
        else:
            raise ValueError(
                'Unknown interpolation mode {}'.format(self._interp_mode))

    def compute_opacity(self, temperature, pressure, wngrid=None):

        return self.interp_bilinear_grid(temperature, pressure, *self.find_closest_index(temperature, pressure), wngrid) / 10000



from .binner import Binner
from taurex.util import bindown
from taurex.util.util import compute_bin_edges, wnwidth_to_wlwidth
from taurex import OutputSize


class SimpleBinner(Binner):
    """
    Bins to a wavenumber grid given by ``wngrid``.
    The method places flux into the correct bins
    using histogramming methods. This is fast but can
    suffer as it assumes that there are no gaps in the
    wavenumber grid. This can cause weird results and
    may cause the flux to be higher in the boundary
    of points between two distinct regions
    (such as WFC3 + Spitzer)

    Parameters
    ----------

    wngrid: :obj:`array`
        Wavenumber grid

    wngrid_width: :obj:`array`, optional
        Must have same shape as ``wngrid``
        Full bin widths for each wavenumber grid point
        given in ``wngrid``. If not provided then
        this is automatically computed from ``wngrid``.

    """

    def __init__(self, wngrid, wngrid_width=None):

        self._wngrid = wngrid
        self._wn_width = wngrid_width
        if self._wn_width is None:
            self._wn_width = compute_bin_edges(self._wngrid)[-1]

    def bindown(self, wngrid, spectrum, grid_width=None, error=None):
        """

        Bins down spectrum.

        Parameters
        ----------
        wngrid : :obj:`array`
            The wavenumber grid of the spectrum to be binned down.

        spectrum: :obj:`array`
            The spectra we wish to bin-down. Must be same shape as
            ``wngrid``.

        grid_width: :obj:`array`, optional
            Wavenumber grid full-widths for the spectrum to be binned down.
            Must be same shape as ``wngrid``.
            Optional.

        error: :obj:`array`, optional
            Associated errors or noise of the spectrum. Must be same shape
            as ``wngrid``.Optional parameter.

        Returns
        -------
        binned_wngrid : :obj:`array`
            New wavenumber grid

        spectrum: :obj:`array`
            Binned spectrum.

        grid_width: :obj:`array`
            New grid-widths

        error: :obj:`array` or None
            Binned error if given else ``None``

        """
        return self._wngrid, bindown(wngrid, spectrum, self._wngrid), \
            None, self._wn_width

    def generate_spectrum_output(self, model_output,
                                 output_size=OutputSize.heavy):

        output = super().generate_spectrum_output(model_output,
                                                  output_size=output_size)
        output['binned_wngrid'] = self._wngrid
        output['binned_wlgrid'] = 10000/self._wngrid
        output['binned_wnwidth'] = self._wn_width
        output['binned_wlwidth'] = wnwidth_to_wlwidth(self._wngrid,
                                                      self._wn_width)
        return output



from .instrument import Instrument
import numpy as np
import math


class SNRInstrument(Instrument):
    """

    Simple instrument model that, for a given
    wavelength-independant, signal-to-noise ratio,
    compute resulting noise from it.

    Parameters
    ----------

    SNR: float
        Signal-to-noise ratio

    binner: :class:`~taurex.binning.binner.Binner`, optional
        Optional resampler to generate a new spectral
        grid.


    """

    def __init__(self, SNR=10, binner=None):
        super().__init__()

        self._binner = binner
        self._SNR = SNR

    def model_noise(self, model, model_res=None, num_observations=1):

        if model_res is None:
            model_res = model.model()

        binner = self._binner
        if binner is None:
            binner = model.defaultBinner()

        wngrid, spectrum, error, grid_width = self._binner.bin_model(model_res)

        signal = spectrum.max() - spectrum.min()

        noise = np.ones(spectrum.shape)*signal/self._SNR

        return wngrid, spectrum, \
            noise / math.sqrt(num_observations), grid_width



from .chemistry import Chemistry
import numpy as np
from taurex.cache import OpacityCache


class ChemistryFile(Chemistry):
    """
    Chemistry profile read from file

    Parameters
    ----------

    gases : :obj:`list`
        Gases in file

    filename : str
        filename for mix ratios


    """

    def __init__(self, gases=None, filename=None):
        super().__init__(self.__class__.__name__)

        self._gases = gases

        self._mix_ratios = np.loadtxt(filename).T
        self.molecules_i_have = OpacityCache().find_list_of_molecules()

        active = []
        active_profile = []
        inactive = []
        inactive_profile = []

        for i, g in enumerate(self._gases):
            if g in self.molecules_i_have:
                active.append(g)
                active_profile.append(self._mix_ratios[i, :])
            else:
                inactive.append(g)
                inactive_profile.append(self._mix_ratios[i, :])

        self.active_gases = active
        self.inactive_gases = inactive

        self.active_mixratio_profile = np.array(active_profile)
        self.inactive_mixratio_profile = np.array(inactive_profile)

    @property
    def activeGases(self):
        """Returns names of actively absorbing molecules

        Returns
        -------

        active : :obj:`list` of str
            List of molecules

        """
        return self.active_gases

    @property
    def inactiveGases(self):
        """Returns names of molecules not actively absorbing

        Returns
        -------

        inactive : :obj:`list` of str
            List of molecules

        """
        return self.inactive_gases

    @property
    def activeGasMixProfile(self):
        """
        Layer by layer mixing ratio of ``active`` molecules

        Returns
        -------

        mix_profile :  :obj:`array`
            Array of shape ``(nactivegases,nlayers)``


        """
        return self.active_mixratio_profile

    @property
    def inactiveGasMixProfile(self):
        """
        Layer by layer mixing ratio of ``inactive`` molecules

        Returns
        -------

        mix_profile :  :obj:`array`
            Array of shape ``(ninactivegases,nlayers)``

        """
        return self.inactive_mixratio_profile

    def write(self, output):
        gas_entry = super().write(output)
        gas_entry.write_scalar('metallicity', self.ace_metallicity)
        gas_entry.write_scalar('co_ratio', self.ace_co)

        return gas_entry
